# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for generic notifications."""

import logging
from typing import ClassVar, override
from unittest import mock

from asgiref.sync import sync_to_async
from django.core import mail
from django.core.mail import EmailMessage

from debusine.db.models import NotificationChannel, WorkRequest, Worker
from debusine.server import notifications
from debusine.tasks.models import (
    ActionSendNotification,
    EventReactions,
    NotificationDataEmail,
)
from debusine.test.django import ChannelsHelpersMixin, TestCase


class NotifyWorkerTokenDisabledTests(ChannelsHelpersMixin, TestCase):
    """Tests for notify_worker_token_disabled."""

    async def test_message_is_not_sent(self) -> None:
        """Test Token with no associated worker: message is not sent."""
        token = await sync_to_async(self.playground.create_user_token)()
        channel = await self.create_channel(token.hash)

        await sync_to_async(token.disable)()

        await self.assert_channel_nothing_received(channel)

    async def test_message_is_sent(self) -> None:
        """Test Token with an associated worker: message is sent."""
        token = await sync_to_async(self.playground.create_worker_token)()
        channel = await self.create_channel(token.hash)

        await sync_to_async(token.disable)()

        await self.assert_channel_received(channel, {"type": "worker.disabled"})


class NotifyWorkRequestAssigned(ChannelsHelpersMixin, TestCase):
    """Tests for notify_work_request_assigned."""

    work_request: ClassVar[WorkRequest]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        workspace = cls.playground.create_workspace(name="Tests")
        token = cls.playground.create_user_token()
        cls.work_request = cls.playground.create_worker_task(
            workspace=workspace, created_by=token.user, mark_pending=True
        )

    def test_message_is_not_sent(self) -> None:
        """Test WorkRequest.assign_worker(None) is a no-operation."""
        # Asserts that no exception is raised. notify_work_request_assigned
        # is checking that worker is None and not trying to access the
        # worker's token.
        self.work_request.assign_worker(None)  # type: ignore[arg-type]

    async def test_message_is_sent(self) -> None:
        """Test WorkRequest.assign_worker(worker) sends msg to the channel."""
        token = await sync_to_async(self.playground.create_worker_token)()
        worker = await sync_to_async(getattr)(token, "worker")
        channel = await self.create_channel(token.hash)

        def assign() -> None:
            with self.captureOnCommitCallbacks(execute=True):
                self.work_request.assign_worker(worker)

        await sync_to_async(assign)()

        await self.assert_channel_received(
            channel, {"type": "work_request.assigned"}
        )

    def test_celery_worker(self) -> None:
        """
        WorkRequest.assign_worker sends no message for Celery workers.

        These are assumed to notify the worker of work to do via some other
        mechanism (e.g. a Celery queue).
        """
        worker = Worker.objects.get_or_create_celery()

        # Asserts that no exception is raised.
        self.work_request.assign_worker(worker)


class NotifyWorkRequestCompletedTests(TestCase):
    """Test notify_work_request_completed function."""

    def assert_send_notification(
        self,
        status: WorkRequest.Results,
        expected_subject: str,
        *,
        expected_to: list[str] | None = None,
        event_name: str = 'on_failure',
        notification_data: NotificationDataEmail | None = None,
        expected_cc: list[str] | None = None,
    ) -> None:
        """Assert WorkRequest completed in failure or error."""
        notification_channel_name = "deblts-email"

        action = ActionSendNotification(channel=notification_channel_name)
        if notification_data is not None:
            action.data = notification_data

        event_reactions = EventReactions()
        setattr(event_reactions, event_name, [action])

        work_request = self.playground.create_worker_task(
            mark_running=True, event_reactions=event_reactions
        )

        if expected_to is None:
            to = ["recipient@example.com"]
        else:
            to = expected_to

        notification_channel_data = {
            "from": "sender@debusine.example.org",
            "to": to,
        }

        NotificationChannel.objects.create(
            name=notification_channel_name,
            method=NotificationChannel.Methods.EMAIL,
            data=notification_channel_data,
        )

        work_request.mark_completed(status)

        self.assertEqual(len(mail.outbox), 1)

        sent_mail = mail.outbox[0]

        self.assertEqual(
            sent_mail.subject,
            expected_subject.format(work_request_id=work_request.id),
        )
        self.assertEqual(sent_mail.body, "Sent by debusine")

        self.assertEqual(sent_mail.to, to)

        if expected_cc is None:
            expected_cc = []

        self.assertEqual(sent_mail.cc, expected_cc)

    def test_send_notification_on_success(self) -> None:
        """One notification email is sent."""
        to = ["a@example.com"]
        cc = ["cc@example.com"]

        self.assert_send_notification(
            WorkRequest.Results.SUCCESS,
            event_name='on_success',
            notification_data=NotificationDataEmail(to=to, cc=cc),
            expected_to=to,
            expected_cc=cc,
            expected_subject=(
                "WorkRequest {work_request_id} completed in success"
            ),
        )

    def test_send_notification_no_to(self) -> None:
        """One notification email is sent."""
        cc = ["cc@example.com"]

        self.assert_send_notification(
            WorkRequest.Results.SUCCESS,
            event_name='on_success',
            notification_data=NotificationDataEmail(cc=cc),
            expected_cc=cc,
            expected_subject=(
                "WorkRequest {work_request_id} completed in success"
            ),
        )

    def test_send_notification_on_failure(self) -> None:
        """
        One notification email is sent.

        The subject {invalid_variable} is not replaced
        """
        to = ["a@example.com"]
        self.assert_send_notification(
            WorkRequest.Results.FAILURE,
            notification_data=NotificationDataEmail(
                to=to,
                subject=(
                    "Bad news: $work_request_id completed "
                    "in $work_request_result $invalid_variable"
                ),
            ),
            expected_to=to,
            expected_subject="Bad news: {work_request_id} completed in failure "
            "$invalid_variable",
        )

    def test_send_notification_on_error(self) -> None:
        """One notification email is sent."""
        to = ["a@example.com"]
        cc = ["cc@example.com"]

        self.assert_send_notification(
            WorkRequest.Results.ERROR,
            notification_data=NotificationDataEmail(to=to, cc=cc),
            expected_to=to,
            expected_cc=cc,
            expected_subject="WorkRequest {work_request_id} completed in error",
        )

    def test_send_notification_on_error_without_data(self) -> None:
        """Mail is sent, WorkRequest notification's data is not provided."""
        self.assert_send_notification(
            WorkRequest.Results.ERROR,
            expected_subject="WorkRequest {work_request_id} completed in error",
        )

    def test_no_notification_is_sent(self) -> None:
        """No mail notification is sent."""
        for result in (
            WorkRequest.Results.SUCCESS,
            WorkRequest.Results.FAILURE,
        ):
            work_request = self.playground.create_worker_task(
                mark_running=True, event_reactions=EventReactions()
            )
            with mock.patch(
                'debusine.db.models.WorkRequest'
                '.process_update_collection_with_artifacts'
            ):
                work_request.mark_completed(result)
                self.assertEqual(len(mail.outbox), 0)

    def test_log_notification_channel_does_not_exist(self) -> None:
        """Log contains debug error that notification channel does not exist."""
        channel_name = "does-not-exist"
        event_reactions = EventReactions(
            on_failure=[
                ActionSendNotification(
                    channel=channel_name,
                    data=NotificationDataEmail(to=["a@example.com"]),
                )
            ]
        )
        work_request = self.playground.create_worker_task(
            mark_running=True, event_reactions=event_reactions
        )

        expected_log = (
            f"WorkRequest {work_request.id}: work-request-completed "
            "notification cannot be sent: "
            f"NotificationChannel {channel_name} does not exist"
        )
        with self.assertLogsContains(
            expected_log, logger="debusine", level=logging.DEBUG
        ):
            work_request.mark_completed(WorkRequest.Results.FAILURE)

    def test_send_work_request_completed_fail_logs(self) -> None:
        """Send email fails: log failure."""
        work_request = self.playground.create_worker_task()
        notification_channel = NotificationChannel.objects.create(
            name="deblts-email",
            method=NotificationChannel.Methods.EMAIL,
            data={
                "from": "sender@debusine.example.org",
                "to": ["rcpt@example.com"],
            },
        )

        work_request_notification = ActionSendNotification(channel="test")
        body = "Not used"

        email_send_patch = mock.patch.object(
            EmailMessage, "send", return_value=0
        )
        mocked = email_send_patch.start()
        self.addCleanup(email_send_patch.stop)

        expected_log = (
            f"NotificationChannel email notification for "
            f"WorkRequest {work_request.id} failed sending"
        )
        with self.assertLogsContains(
            expected_log, logger="debusine", level=logging.WARNING
        ):
            notifications._send_work_request_completed_notification(
                notification_channel,
                work_request,
                work_request_notification,
                body,
            )

        mocked.assert_called_once_with(fail_silently=True)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine auth views."""

import abc
from collections import defaultdict
from functools import cached_property
from typing import Any, cast, override

from django.contrib.auth import views as auth_views
from django.contrib.auth.forms import AuthenticationForm
from django.db.models import QuerySet
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse

from debusine.db.context import context
from debusine.db.models import Group, Scope, Token, User
from debusine.db.models.auth import (
    GroupAuditLog,
    GroupMembership,
    GroupQuerySet,
    Identity,
    UserQuerySet,
)
from debusine.server.signon.providers import Provider
from debusine.server.signon.signon import Signon
from debusine.server.signon.views import SignonLogoutMixin
from debusine.web.forms import (
    BootstrapMixin,
    GroupAddUserForm,
    ModelFormBase,
    TokenForm,
)
from debusine.web.views.base import (
    BaseUIView,
    CreateViewBase,
    DeleteViewBase,
    DetailViewBase,
    FormMixinBase,
    ListViewBase,
    SingleObjectMixinBase,
    UpdateViewBase,
)
from debusine.web.views.places import IndexPlace, Place, ResourcePlace
from debusine.web.views.table import TableMixin
from debusine.web.views.tables import (
    GroupAuditLogTable,
    GroupMembershipAdminTable,
    GroupMembershipTable,
)


class AuthenticationBootstrapForm(BootstrapMixin, AuthenticationForm):
    """Django's AuthenticationForm using BootstrapMixin from Debusine."""


class LoginView(BaseUIView, auth_views.LoginView):
    """Class for the login view."""

    base_template = "web/_base_debusine.html"
    template_name = "account/login.html"
    form_class = AuthenticationBootstrapForm
    title = "Log in"

    @override
    def get(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponse:
        """Store the ?sso_callback_next_url argument in the session."""
        request.session["sso_callback_next_url"] = self.get_success_url()
        return super().get(request, *args, **kwargs)

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Tell the base template we are a login view."""
        ctx = super().get_context_data(**kwargs)
        ctx["is_login_view"] = True
        return ctx


class LogoutView(
    auth_views.LogoutView,
    SignonLogoutMixin,
    BaseUIView,
):
    """Class for the logout view."""

    base_template = "web/_base_debusine.html"
    template_name = "account/logged_out.html"
    title = "Logged out"

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Tell the base template we are a logout view."""
        ctx = super().get_context_data(**kwargs)
        ctx["is_logout_view"] = True
        return ctx


class UserDetailView(BaseUIView, DetailViewBase[User]):
    """Show user details."""

    model = User
    base_template = "web/_base_debusine.html"
    template_name = "web/user-detail.html"
    # We need something that is not "user" not to confuse it with the
    # current user
    context_object_name = "person"

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place

    @override
    def get_queryset(self) -> QuerySet[User]:
        """All workspaces for authenticated users or only public ones."""
        queryset = cast(UserQuerySet[Any], super().get_queryset())
        return queryset.can_display(context.pc)

    @override
    def get_object(self, queryset: QuerySet[User] | None = None) -> User:
        """Get the User to show."""
        assert queryset is None
        queryset = self.get_queryset()
        return get_object_or_404(queryset, username=self.kwargs["username"])

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Add user information to context data."""
        ctx = super().get_context_data(**kwargs)

        # Groups by scope
        by_group: dict[Scope, list[Group]] = defaultdict(list)
        for group in self.object.debusine_groups.all():
            by_group[group.scope].append(group)
        ctx["groups_by_scope"] = sorted(
            (
                (scope, sorted(groups, key=lambda g: g.name))
                for scope, groups in by_group.items()
            ),
            key=lambda s: s[0].name,
        )

        if self.object.can_manage(context.pc):
            # SSO identities
            signon: Signon | None = getattr(self.request, "signon", None)
            if signon is not None:
                identities: list[tuple[Provider, Identity]] = []
                for identity in self.object.identities.all():
                    if (
                        provider := signon.get_provider_for_identity(identity)
                    ) is None:
                        continue
                    identities.append((provider, identity))
                ctx["identities"] = identities

        return ctx


class UserIDRedirectView(BaseUIView, DetailViewBase[User]):
    """Redirect to a UserDetailView."""

    model = User
    context_object_name = "person"

    @override
    def get_queryset(self) -> QuerySet[User]:
        """All workspaces for authenticated users or only public ones."""
        queryset = cast(UserQuerySet[Any], super().get_queryset())
        return queryset.can_display(context.pc)

    @override
    def get(
        self,
        request: HttpRequest,  # noqa: ARG002
        *args: Any,  # noqa: ARG002
        **kwargs: Any,  # noqa: ARG002
    ) -> HttpResponse:
        """Redirect to UserDetailView."""
        return HttpResponseRedirect(self.get_object().get_absolute_url())


class TokenView(BaseUIView):
    """Base view for token management."""

    base_template = "web/_base_debusine.html"

    @cached_property
    def user(self) -> User:
        """Return the user for which we manage tokens."""
        user = get_object_or_404(
            User.objects.can_display(context.pc),
            username=self.kwargs["username"],
        )
        self.enforce(user.can_manage)
        return user

    @override
    def get_place(self) -> Place:
        return IndexPlace(
            title="User tokens",
            description=f"Token list for {self.user}",
            url=self.request.get_full_path(),
            parent=self.user.ui(self.request).place,
        )

    def get_success_url(self) -> str:
        """Redirect to the view to see the created token."""
        return reverse(
            "user:token-list", kwargs={"username": self.user.username}
        )

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Add the currently managed user to the template context."""
        ctx = super().get_context_data(**kwargs)
        ctx["person"] = self.user
        return ctx


class UserTokenListView(TokenView, ListViewBase[Token]):
    """List tokens for the user."""

    model = Token
    template_name = "web/user_token-list.html"
    context_object_name = "token_list"
    ordering = "name"
    title = "API tokens"

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Add bits needed by the template."""
        ctx = super().get_context_data(**kwargs)
        ctx["only_one_scope"] = len(ctx["scopes"]) == 1
        return ctx

    @override
    def get_queryset(self) -> QuerySet[Token]:
        """All tokens for the authenticated user."""
        return Token.objects.filter(user=self.user).order_by('created_at')


class TokenFormView(TokenView, FormMixinBase[TokenForm], abc.ABC):
    """Mixin for views using TokenForm."""

    form_class = TokenForm
    action: str

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Add action name to context data."""
        ctx = super().get_context_data(**kwargs)
        ctx["action"] = self.action
        return ctx

    @override
    def get_form_kwargs(self) -> dict[str, Any]:
        """Extend the default kwarg arguments: add "user"."""
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.user
        return kwargs


class UserTokenCreateView(TokenFormView, CreateViewBase[Token, TokenForm]):
    """Form view for creating tokens."""

    template_name = "web/user_token-form.html"
    action = "Create"
    context_object_name = "token"

    @override
    def get_form_kwargs(self) -> dict[str, Any]:
        """Extend the default kwarg arguments: add "user"."""
        kwargs = super().get_form_kwargs()
        kwargs["user"] = self.user
        return kwargs

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Add the user object from URL kwargs."""
        ctx = super().get_context_data(**kwargs)
        ctx["person"] = self.user
        return ctx

    @override
    def get_place(self) -> Place:
        if self.object:
            title = "Token created"
        else:
            title = f"{self.action} token"

        return IndexPlace(
            title=title,
            url=reverse(
                "user:token-create", kwargs={"username": self.user.username}
            ),
            parent=self.user.ui(self.request).place,
        )

    @override
    def get_template_names(self) -> list[str]:
        """Switch template if we have a newly created token."""
        if self.object:
            return ["web/user_token-created.html"]
        else:
            return [self.template_name]

    @override
    def form_valid(self, form: TokenForm) -> HttpResponse:
        """Validate form and return token created page."""
        # FIXME: a successful form submission should not render a page and
        # should redirect to another page instead, to avoid unintended
        # behaviour if the user triggers a reload. See:
        # https://developer.mozilla.org/en-US/docs/Web/HTTP/Guides/Redirections#temporary_responses_to_unsafe_requests # noqa: E501
        #
        # However, here we lose the token key after we send a response, so to
        # avoid rendering a page to show it here, we need to store it somewhere
        # the target of the redirect can find it.
        #
        # That would introduce further issues in managing the security and the
        # lifetime of that storage, and the worst that can happen with a reload
        # here is the generation of another token, which seems like the lesser
        # risk for now.
        self.object = form.save()
        return self.render_to_response(self.get_context_data(**self.kwargs))


class UserTokenUpdateView(TokenFormView, UpdateViewBase[Token, TokenForm]):
    """Form view for creating tokens."""

    model = Token
    template_name = "web/user_token-form.html"
    action = "Edit"

    @override
    def get_place(self) -> Place:
        return IndexPlace(
            title=f"{self.action} token",
            url=reverse(
                "user:token-edit",
                kwargs={"username": self.user.username, "pk": self.object.pk},
            ),
            parent=self.user.ui(self.request).place,
        )

    @override
    def get_queryset(self) -> QuerySet[Token]:
        """Only include tokens for the current user."""
        return super().get_queryset().filter(user=self.user)


class UserTokenDeleteView(
    TokenView, DeleteViewBase[Token, ModelFormBase[Token]]
):
    """View for deleting tokens."""

    object: Token  # https://github.com/typeddjango/django-stubs/issues/1227
    model = Token
    template_name = "web/user_token-confirm_delete.html"
    title = "Delete token"

    @override
    def get_queryset(self) -> QuerySet[Token]:
        """Only include tokens for the current user."""
        return super().get_queryset().filter(user=self.user)


class UserGroupListView(ListViewBase[GroupMembership], BaseUIView):
    """List groups for a user."""

    model = GroupMembership
    base_template = "web/_base_debusine.html"
    template_name = "web/user_group-list.html"
    context_object_name = "membership_list"
    ordering = "group__name"

    @override
    def get_place(self) -> Place:
        return context.require_scope().ui(self.request).place_group_list

    @cached_property
    def user(self) -> User:
        """Return the user for which we list/manage groups."""
        user = get_object_or_404(
            User.objects.all(),
            username=self.kwargs["username"],
        )
        self.enforce(user.can_display)
        return user

    @override
    def get_queryset(self) -> QuerySet[GroupMembership]:
        """All groups for the authenticated user."""
        return GroupMembership.objects.filter(
            group__in=Group.objects.can_display(context.pc),
            group__scope=context.require_scope(),
            user=self.user,
        ).select_related("group", "user")

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Add bits needed by the template."""
        ctx = super().get_context_data(**kwargs)
        ctx["current_user"] = self.request.user == self.user
        ctx["person"] = self.user
        return ctx


class BaseGroupUIView(BaseUIView):
    """Base class for group UI views."""

    @override
    def init_view(self) -> None:
        """Set the current workspace, if given."""
        super().init_view()
        if "wname" in self.kwargs:
            self.set_current_workspace(self.kwargs["wname"])


class GroupDetailView(
    FormMixinBase[GroupAddUserForm],
    DetailViewBase[Group],
    BaseGroupUIView,
):
    """Show details about a group."""

    model = Group
    form_class = GroupAddUserForm
    template_name = "web/group-detail.html"
    context_object_name = "group"

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place

    @override
    def get_object(
        self, queryset: QuerySet[Group, Group] | None = None
    ) -> Group:
        """Get the Group to display."""
        assert queryset is None
        queryset = self.get_queryset()
        group = get_object_or_404(
            cast(GroupQuerySet[Any], queryset).can_display(context.pc),
            scope=context.require_scope(),
            workspace=context.workspace,
            name=self.kwargs["group"],
        )
        return cast(Group, group)

    @override
    def get_form_kwargs(self) -> dict[str, Any]:
        """Extend the default kwarg arguments: add "group"."""
        kwargs = super().get_form_kwargs()
        kwargs["group"] = self.object
        return kwargs

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Add bits needed by the template."""
        ctx = super().get_context_data(
            asc=self.request.GET.get("asc", "1"),
            order=self.request.GET.get("order"),
            **kwargs,
        )

        ctx["can_manage"] = can_manage = self.object.can_manage(context.pc)

        table_class = (
            GroupMembershipAdminTable if can_manage else GroupMembershipTable
        )
        ctx["members"] = table_class(
            request=self.request,
            object_list=GroupMembership.objects.filter(
                group=self.object
            ).select_related("user"),
        ).get_paginator(per_page=10)

        audit_log = self.object.audit_log.order_by(
            "-created_at"
        ).select_related("actor")[:6]
        ctx["audit_log"] = audit_log[:5]
        ctx["audit_log_more"] = len(audit_log) == 6
        # This is enforced by GroupQuerySet.can_display called in
        # self.get_object
        assert self.request.user.is_authenticated
        try:
            ctx["role"] = GroupMembership.objects.get(
                group=self.object, user=self.request.user
            ).role
        except GroupMembership.DoesNotExist:
            pass

        return ctx

    # Adapted from BaseDeleteView, since ProcessFormView does not set
    # self.object
    def post(
        self,
        request: HttpRequest,  # noqa: ARG002, U100
        *args: Any,
        **kwargs: Any,
    ) -> HttpResponse:
        """Handle POST requests."""
        self.object = self.get_object()
        self.enforce(self.object.can_manage)
        form = self.get_form()
        if form.is_valid():
            return self.form_valid(form)
        else:
            return self.form_invalid(form)

    @override
    def form_valid(self, form: GroupAddUserForm) -> HttpResponse:
        """Validate form and return user created page."""
        username = form.cleaned_data["username"]
        user = User.objects.get(username=username)
        self.object.add_user(user)
        return super().form_valid(form)

    @override
    def get_success_url(self) -> str:
        """Redirect to the view to see the created user."""
        return self.object.get_absolute_url()


class MembershipView(SingleObjectMixinBase[GroupMembership], BaseGroupUIView):
    """Base view for group management for a user."""

    @override
    def get_object(
        self, queryset: QuerySet[GroupMembership, GroupMembership] | None = None
    ) -> GroupMembership:
        """Get the GroupMembership to display."""
        assert queryset is None
        queryset = self.get_queryset()
        membership = get_object_or_404(
            queryset.filter(
                group__in=Group.objects.can_manage(context.pc),
                group__scope=context.require_scope(),
                group__name=self.kwargs["group"],
                group__workspace=context.workspace,
                user__username=self.kwargs["user"],
            ).select_related("group", "user"),
        )
        self.enforce(membership.group.can_manage)
        return membership

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Add bits needed by the template."""
        ctx = super().get_context_data(**kwargs)
        # FIXME: Django defines self.object differently in separate
        # incompatible view classes, and I could not find a way to play along
        # well with that
        ctx["person"] = self.object.user  # type: ignore[attr-defined]
        ctx["group"] = self.object.group  # type: ignore[attr-defined]
        return ctx


class MembershipUpdateView(
    MembershipView,
    UpdateViewBase[GroupMembership, ModelFormBase[GroupMembership]],
):
    """Edit a user's membership."""

    model = GroupMembership
    fields = ["role"]
    template_name = "web/membership-edit.html"
    title = "Update role"

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place_update

    @override
    def form_valid(self, form: ModelFormBase[GroupMembership]) -> HttpResponse:
        """Change the member's role."""
        self.object = self.object.group.set_user_role(
            self.object.user, form.cleaned_data["role"]
        )
        return HttpResponseRedirect(self.get_success_url())

    @override
    def get_success_url(self) -> str:
        """Redirect to the view to see the updated group."""
        return self.object.group.get_absolute_url()


class MembershipDeleteView(
    MembershipView,
    DeleteViewBase[GroupMembership, ModelFormBase[GroupMembership]],
):
    """View for removing a user from a group."""

    # https://github.com/typeddjango/django-stubs/issues/1227
    object: GroupMembership
    model = GroupMembership
    template_name = "web/membership-confirm_delete.html"

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place_remove

    @override
    def form_valid(
        self,
        form: ModelFormBase[GroupMembership],  # noqa: ARG002, U100
    ) -> HttpResponse:
        """Remove the group member."""
        success_url = self.get_success_url()
        self.object.group.remove_user(self.object.user)
        return HttpResponseRedirect(success_url)

    @override
    def get_success_url(self) -> str:
        """Redirect to the view to see the updated group."""
        return self.object.group.get_absolute_url()


class GroupAuditLogView(
    TableMixin[GroupAuditLog], ListViewBase[GroupAuditLog], BaseGroupUIView
):
    """View for showing the full group audit log."""

    model = GroupAuditLog
    template_name = "web/group_audit_log-list.html"
    context_object_name = "audit_log"
    table_class = GroupAuditLogTable
    paginate_by = 10

    @cached_property
    def group(self) -> Group:
        """Return the group for the audit log."""
        return cast(
            Group,
            get_object_or_404(
                cast(
                    GroupQuerySet[Any],
                    Group.objects.can_display_audit_log(context.pc).filter(
                        scope=context.require_scope()
                    ),
                ),
                name=self.kwargs["group"],
                workspace=context.workspace,
            ),
        )

    @override
    def get_queryset(self) -> QuerySet[GroupAuditLog]:
        """All audit log entries for the current group."""
        return self.group.audit_log.select_related("actor")

    @override
    def get_place(self) -> ResourcePlace:
        return self.group.ui(self.request).place_audit_log

    @override
    def get_context_data(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        """Add bits needed by the template."""
        ctx = super().get_context_data(**kwargs)
        ctx["group"] = self.group
        return ctx



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for override_query_parameters."""

from unittest import TestCase
from unittest.mock import MagicMock

from django.http import QueryDict

from debusine.web.templatetags.override_query_parameters import (
    override_query_parameters,
)


class OverrideQueryParametersTests(TestCase):
    """Tests for override_query_parameters() function."""

    def test_override_query_parameters(self) -> None:
        """Test override_query_parameters() return modified query parameters."""
        context = MagicMock()
        context.request.GET = QueryDict(query_string="page=1&order=id&asc=1")

        page = 2
        result = override_query_parameters(context, page=page)

        self.assertEqual(result, f"page={page}&order=id&asc=1")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for db artifacts."""

import datetime as dt
from functools import partial
from pathlib import Path
from typing import Any, Optional, Self, TYPE_CHECKING, TypedDict, override

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import PermissionDenied, ValidationError
from django.db import models, transaction
from django.db.models import Exists, F, OuterRef, Q, QuerySet, UniqueConstraint
from django.db.models.functions import Coalesce
from django.urls import reverse
from django_cte import CTE, with_cte

from debusine.artifacts import LocalArtifact
from debusine.artifacts.models import ArtifactCategory, ArtifactData
from debusine.db.models.files import File
from debusine.db.models.permissions import (
    Allow,
    permission_check,
    permission_filter,
)
from debusine.db.models.workspaces import Workspace
from debusine.db.permissioncontext import PermissionContext
from debusine.utils.typing_utils import copy_signature_from

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django_stubs_ext import WithAnnotations
    from django_stubs_ext.db.models import TypedModelMeta

    from debusine.db.models.work_requests import WorkRequest
    from debusine.web.views.ui.artifacts import ArtifactUI, FileInArtifactUI

    ArtifactWithComplete = WithAnnotations["Artifact", "CompleteDict"]

    WithAnnotations  # fake usage for vulture

else:
    TypedModelMeta = object

ARTIFACT_CATEGORY_ICON_NAMES = {
    ArtifactCategory.AUTOPKGTEST: "folder",
    ArtifactCategory.BINARY_PACKAGE: "folder",
    ArtifactCategory.BINARY_PACKAGES: "folder",
    ArtifactCategory.BLHC: "folder",
    ArtifactCategory.DEBDIFF: "folder",
    ArtifactCategory.LINTIAN: "folder",
    ArtifactCategory.PACKAGE_BUILD_LOG: "journal-text",
    ArtifactCategory.SIGNING_INPUT: "folder",
    ArtifactCategory.SIGNING_OUTPUT: "folder",
    ArtifactCategory.SOURCE_PACKAGE: "folder",
    ArtifactCategory.SYSTEM_IMAGE: "folder",
    ArtifactCategory.SYSTEM_TARBALL: "folder",
    ArtifactCategory.UPLOAD: "folder",
    ArtifactCategory.WORK_REQUEST_DEBUG_LOGS: "folder",
}

ARTIFACT_CATEGORY_SHORT_NAMES = {
    ArtifactCategory.AUTOPKGTEST: "autopkgtest",
    ArtifactCategory.BINARY_PACKAGE: "binary package",
    ArtifactCategory.BINARY_PACKAGES: "binary packages",
    ArtifactCategory.BLHC: "blhc report",
    ArtifactCategory.DEBDIFF: "debdiff report",
    ArtifactCategory.LINTIAN: "lintian report",
    ArtifactCategory.PACKAGE_BUILD_LOG: "build log",
    ArtifactCategory.SIGNING_INPUT: "signing input",
    ArtifactCategory.SIGNING_OUTPUT: "signing output",
    ArtifactCategory.SOURCE_PACKAGE: "source package",
    ArtifactCategory.SYSTEM_IMAGE: "system image",
    ArtifactCategory.SYSTEM_TARBALL: "system tar",
    ArtifactCategory.UPLOAD: "package upload",
    ArtifactCategory.WORK_REQUEST_DEBUG_LOGS: "debug log",
}


class CompleteDict(TypedDict):
    """Additional fields returned by ArtifactQuerySet.annotate_complete."""

    complete: bool


class ArtifactQuerySet[A](QuerySet["Artifact", A]):
    """Custom QuerySet for Artifact."""

    def in_current_scope(self) -> Self:
        """Filter to artifacts in the current scope."""
        from debusine.db.context import context

        return self.filter(workspace__scope=context.require_scope())

    def in_current_workspace(self) -> Self:
        """Filter to artifacts in the current workspace."""
        from debusine.db.context import context

        return self.filter(workspace=context.require_workspace())

    def visible_to_current_workspace(self) -> Self:
        """Filter to artifacts that the current workspace inherits from."""
        from debusine.db.context import context

        visible = CTE(
            Workspace.objects.can_display(context.pc),
            name="visible",
        )
        chain = CTE.recursive(
            context.require_workspace().make_inheritance_chain_cte(visible),
            name="chain",
        )
        return self.filter(
            workspace__in=with_cte(
                visible,
                chain,
                select=chain.join(Workspace, pk=chain.col.chain_parent),
            )
        )

    def work_request_can_display(self, work_request: "WorkRequest") -> Self:
        """Keep only artifacts that the work request can display."""
        task = work_request.get_task()
        return self.filter(
            Q(pk__in=task.get_input_artifacts_ids())
            | Q(created_by_work_request=work_request)
        )

    def work_request_can_change(self, work_request: "WorkRequest") -> Self:
        """Keep only artifacts that the work request can change."""
        return self.filter(created_by_work_request=work_request)

    @permission_filter(work_request=Allow.PASS, anonymous=Allow.PASS)
    def can_display(self, pc: PermissionContext) -> Self:
        """Keep only Artifacts that can be displayed."""
        # Delegate to workspace can_display check
        qs = self.filter(workspace__in=Workspace.objects.can_display(pc))
        if pc.work_request is not None:
            qs = qs.work_request_can_display(pc.work_request)
        return qs

    @permission_filter(work_request=Allow.PASS)
    def can_add_files(self, pc: PermissionContext) -> Self:
        """Keep only Artifacts where files can be added."""
        qs = self.filter(
            workspace__in=Workspace.objects.can_create_artifacts(pc)
        )
        if pc.work_request is not None:
            qs = qs.work_request_can_change(pc.work_request)
        return qs

    @permission_filter(work_request=Allow.PASS)
    def can_add_relation(self, pc: PermissionContext) -> Self:
        """Keep only Artifacts where relations can be added."""
        qs = self.filter(
            workspace__in=Workspace.objects.can_create_artifacts(pc)
        )
        if pc.work_request is not None:
            qs = qs.work_request_can_change(pc.work_request)
        return qs

    def annotate_complete(
        self,
    ) -> "ArtifactQuerySet[ArtifactWithComplete]":
        """Annotate artifacts with whether all their files are complete."""
        return self.annotate(
            complete=~Exists(
                FileInArtifact.objects.filter(
                    artifact=OuterRef("pk"), complete=False
                )
            )
        )

    def not_expired(self, at: dt.datetime) -> Self:
        """
        Return queryset with artifacts that have not expired.

        :param at: datetime to check if the artifacts are not expired.
        :return: artifacts that expire_at is None (do not expire) or
          expire_at is after the given datetime.
        """
        return self.annotate(
            _effective_expiration_delay=Coalesce(
                "expiration_delay",
                "workspace__default_expiration_delay",
            )
        ).filter(
            Q(_effective_expiration_delay=dt.timedelta(0))
            | Q(
                # https://github.com/typeddjango/django-stubs/issues/1548
                created_at__gt=(
                    at - F("_effective_expiration_delay")  # type: ignore[operator] # noqa: E501
                )
            )
        )

    def expired(self, at: dt.datetime) -> Self:
        """
        Return queryset with artifacts that have expired.

        :param at: datetime to check if the artifacts are expired.
        :return: artifacts that expire_at is before the given datetime.
        """
        return (
            self.annotate(
                _effective_expiration_delay=Coalesce(
                    "expiration_delay",
                    "workspace__default_expiration_delay",
                )
            )
            .exclude(_effective_expiration_delay=dt.timedelta(0))
            .filter(
                # https://github.com/typeddjango/django-stubs/issues/1548
                created_at__lte=(
                    at - F("_effective_expiration_delay")  # type: ignore[operator] # noqa: E501
                )
            )
        )

    def part_of_collection_with_retains_artifacts(self) -> Self:
        """
        Return Artifacts where the parent_collections has retains_artifacts set.

        :return: Artifacts that are part of a retains_artifacts collection.
        """
        # Import here to prevent circular imports
        from debusine.db.models.collections import Collection
        from debusine.db.models.work_requests import WorkRequest

        RetainsArtifacts = Collection.RetainsArtifacts
        return self.filter(
            Q(parent_collections__retains_artifacts=RetainsArtifacts.ALWAYS)
            | Q(
                parent_collections__retains_artifacts=RetainsArtifacts.WORKFLOW,
                parent_collections__workflow__status__in={
                    WorkRequest.Statuses.PENDING,
                    WorkRequest.Statuses.RUNNING,
                    WorkRequest.Statuses.BLOCKED,
                },
            )
        )


class ArtifactManager(models.Manager["Artifact"]):
    """Manager for the Artifact model."""

    @override
    def get_queryset(self) -> ArtifactQuerySet[Any]:
        """Use the custom QuerySet."""
        return ArtifactQuerySet(self.model, using=self._db)

    @classmethod
    def create_from_local_artifact(
        cls,
        local_artifact: LocalArtifact[Any],
        workspace: Workspace,
        *,
        created_by_work_request: Optional["WorkRequest"] = None,
    ) -> "Artifact":
        """Return a new Artifact based on a :py:class:`LocalArtifact`."""
        artifact = Artifact.objects.create(
            category=local_artifact.category,
            workspace=workspace,
            data=local_artifact.data.model_dump(),
            created_by_work_request=created_by_work_request,
        )

        for artifact_path, local_path in local_artifact.files.items():
            file = File.from_local_path(local_path)
            file_backend = workspace.scope.upload_file_backend(file)
            file_backend.add_file(local_path, fileobj=file)
            FileInArtifact.objects.create(
                artifact=artifact,
                path=artifact_path,
                file=file,
                complete=True,
                content_type=local_artifact.content_types.get(artifact_path),
            )

        return artifact


class Artifact(models.Model):
    """Artifact model."""

    category = models.CharField(max_length=255)
    workspace = models.ForeignKey(Workspace, on_delete=models.PROTECT)
    files = models.ManyToManyField(File, through="db.FileInArtifact")
    data = models.JSONField(default=dict, blank=True)
    created_at = models.DateTimeField(auto_now_add=True)
    expiration_delay = models.DurationField(blank=True, null=True)
    created_by = models.ForeignKey(
        "User", blank=True, null=True, on_delete=models.PROTECT
    )
    created_by_work_request = models.ForeignKey(
        "WorkRequest", blank=True, null=True, on_delete=models.SET_NULL
    )
    original_artifact = models.ForeignKey(
        "Artifact", blank=True, null=True, on_delete=models.SET_NULL
    )

    objects = ArtifactManager.from_queryset(ArtifactQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        indexes = [
            # Used by debusine.server.open_metrics.DebusineCollector.  This
            # is imperfect (on debusine.debian.net at the time of writing it
            # only takes the artifact-counting query from ~550ms to ~300ms).
            # Doing any better seems likely to involve a separate
            # trigger-maintained statistics table.
            models.Index(
                "workspace",
                "category",
                name="%(app_label)s_%(class)s_workspace_category",
                include=["id"],
            )
        ]

    @override
    def __str__(self) -> str:
        """Return basic information of Artifact."""
        return (
            f"Id: {self.id} "
            f"Category: {self.category} "
            f"Workspace: {self.workspace.id}"
        )

    @override
    @copy_signature_from(models.Model.save)
    def save(self, **kwargs: Any) -> None:
        """Wrap save with permission checks."""
        from debusine.db.context import context

        if self._state.adding:
            # Create
            if not self.workspace.can_create_artifacts(context.pc):
                user = context.user or AnonymousUser()
                raise PermissionDenied(
                    f"{user} cannot create artifacts in {self.workspace}"
                )
        else:
            # Update
            ...  # TODO: check for update permissions

        return super().save(**kwargs)

    def get_absolute_url(self) -> str:
        """Return the canonical URL to display the artifact."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.workspace.scope.name):
            return reverse(
                "workspaces:artifacts:detail",
                kwargs={"wname": self.workspace.name, "artifact_id": self.id},
            )

    def get_absolute_url_download(self) -> str:
        """Return the canonical URL to download the artifact."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.workspace.scope.name):
            return reverse(
                "workspaces:artifacts:download",
                kwargs={"wname": self.workspace.name, "artifact_id": self.id},
            )

    @override
    def clean(self) -> None:
        """
        Ensure that data is valid for this artifact category.

        :raise ValidationError: for invalid data.
        """
        if not isinstance(self.data, dict):
            raise ValidationError({"data": "data must be a dictionary"})

        try:
            artifact_cls = LocalArtifact.class_from_category(self.category)
        except ValueError as e:
            raise ValidationError(
                {"category": f"{self.category}: invalid artifact category"}
            ) from e

        try:
            artifact_cls.create_data(self.data)
        except ValueError as e:
            raise ValidationError(
                {"data": f"invalid artifact data: {e}"}
            ) from e

    def work_request_can_display(self, work_request: "WorkRequest") -> bool:
        """Check if the work request can display this artifact."""
        if work_request == self.created_by_work_request:
            return True
        task = work_request.get_task()
        return self.pk in task.get_input_artifacts_ids()

    def work_request_can_change(self, work_request: "WorkRequest") -> bool:
        """Check if the work request can change this artifact."""
        return work_request == self.created_by_work_request

    @permission_check(
        "{user} cannot display artifact {resource.id}",
        work_request=Allow.PASS,
        anonymous=Allow.PASS,
    )
    def can_display(self, pc: PermissionContext) -> bool:
        """Check if the artifact can be displayed."""
        if not self.workspace.can_display(pc):
            return False
        if pc.work_request is None:
            return True
        return self.work_request_can_display(pc.work_request)

    @permission_check(
        "{user} cannot add files to artifact {resource.id}",
        work_request=Allow.PASS,
    )
    def can_add_files(self, pc: PermissionContext) -> bool:
        """Check if files can be added to the artifact."""
        if not self.workspace.can_create_artifacts(pc):
            return False
        if pc.work_request is None:
            return True
        return self.work_request_can_change(pc.work_request)

    @permission_check(
        "{user} cannot add relations to artifact {resource.id}",
        work_request=Allow.PASS,
    )
    def can_add_relation(self, pc: PermissionContext) -> bool:
        """Check if the artifact can be displayed."""
        if not self.workspace.can_create_artifacts(pc):
            return False
        if pc.work_request is None:
            return True
        return self.work_request_can_change(pc.work_request)

    def create_data(self) -> ArtifactData:
        """Instantiate ArtifactData from data."""
        artifact_cls = LocalArtifact.class_from_category(self.category)
        artifact_data = artifact_cls.create_data(self.data)
        assert isinstance(artifact_data, ArtifactData)
        return artifact_data

    def get_label(self, data: ArtifactData | None = None) -> str:
        """
        Return a label for this artifact.

        Optionally reuse an already instantiated data model.
        """
        if data is None:
            data = self.create_data()
        if label := data.get_label():
            return label
        return str(self.category)

    def effective_expiration_delay(self) -> dt.timedelta:
        """Return expiration_delay, inherited if None."""
        expiration_delay = self.expiration_delay
        if self.expiration_delay is None:  # inherit
            expiration_delay = self.workspace.default_expiration_delay
        assert expiration_delay is not None
        return expiration_delay

    @property
    def expire_at(self) -> dt.datetime | None:
        """Return computed expiration date."""
        delay = self.effective_expiration_delay()
        if delay == dt.timedelta(0):
            return None
        return self.created_at + delay

    def expired(self, at: dt.datetime) -> bool:
        """
        Return True if this artifact has expired at a given datetime.

        :param at: datetime to check if the artifact is expired.
        :return bool: True if the artifact's expire_at is on or earlier than
          the parameter at.
        """
        expire_at = self.expire_at
        if expire_at is None:
            return False
        return expire_at <= at

    def ui(self, request: "HttpRequest") -> "ArtifactUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.artifacts import ArtifactUI

        return ArtifactUI.get(request, self)


class FileInArtifact(models.Model):
    """File in artifact."""

    artifact = models.ForeignKey(Artifact, on_delete=models.PROTECT)
    path = models.CharField(max_length=500)
    file = models.ForeignKey(File, on_delete=models.PROTECT)
    # We do a best-effort data migration here, but note that in cases where
    # a file store is shared between multiple workspaces, this field may be
    # left as False for files that were uploaded before this field was
    # added.  (debusine.debian.net only had a single workspace at the time
    # of this migration, so that's not a problem there.)
    complete = models.BooleanField(default=False)
    # TODO: ruff is correct that we shouldn't use null=True on a CharField,
    # but fixing that retroactively is non-trivial.
    content_type = models.CharField(blank=True, null=True)  # noqa: DJ001

    class Meta(TypedModelMeta):
        constraints = [
            UniqueConstraint(
                fields=["artifact", "path"],
                name="%(app_label)s_%(class)s_unique_artifact_path",
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return basic information of FileInArtifact."""
        return (
            f"Id: {self.id} Artifact: {self.artifact.id} "
            f"Path: {self.path} File: {self.file.id}"
        )

    def get_absolute_url(self) -> str:
        """Return an absolute URL to view this file."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.artifact.workspace.scope.name):
            return reverse(
                "workspaces:artifacts:file-detail",
                kwargs={
                    "wname": self.artifact.workspace.name,
                    "artifact_id": self.artifact.id,
                    "path": self.path,
                },
            )

    def get_absolute_url_raw(self) -> str:
        """Return an absolute URL to view this file as raw."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.artifact.workspace.scope.name):
            return reverse(
                "workspaces:artifacts:file-raw",
                kwargs={
                    "wname": self.artifact.workspace.name,
                    "artifact_id": self.artifact.id,
                    "path": self.path,
                },
            )

    def get_absolute_url_download(self) -> str:
        """Return an absolute URL to download this file."""
        from debusine.server.scopes import urlconf_scope

        with urlconf_scope(self.artifact.workspace.scope.name):
            return reverse(
                "workspaces:artifacts:file-download",
                kwargs={
                    "wname": self.artifact.workspace.name,
                    "artifact_id": self.artifact.id,
                    "path": self.path,
                },
            )

    def ui(self, request: "HttpRequest") -> "FileInArtifactUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.artifacts import FileInArtifactUI

        return FileInArtifactUI.get(request, self)


class FileUpload(models.Model):
    """File that is being/has been uploaded."""

    file_in_artifact = models.OneToOneField(
        FileInArtifact, on_delete=models.PROTECT
    )
    path = models.CharField(
        max_length=500,
        help_text="Path in the uploads directory",
        unique=True,
    )
    last_activity_at = models.DateTimeField(auto_now_add=True)

    @override
    def __str__(self) -> str:
        """Return basic information."""
        return f"{self.id}"

    @classmethod
    def current_size(cls, artifact: Artifact, path_in_artifact: str) -> int:
        """
        Return current file size.

        The current file size might be smaller than the expected size of the
        file if the file has not finished being uploaded.

        Raise ValueError if path_in_artifact does not exist in Artifact or
        if there's no FileUpload object for the specific File.
        """
        try:
            file_in_artifact = FileInArtifact.objects.get(
                artifact=artifact, path=path_in_artifact
            )
        except FileInArtifact.DoesNotExist:
            raise ValueError(
                f'No FileInArtifact for Artifact {artifact.id} '
                f'and path "{path_in_artifact}"'
            )

        try:
            file_upload = FileUpload.objects.get(
                file_in_artifact=file_in_artifact
            )
        except FileUpload.DoesNotExist:
            raise ValueError(
                f"No FileUpload for FileInArtifact {file_in_artifact.id}"
            )

        try:
            size = file_upload.absolute_file_path().stat().st_size
        except FileNotFoundError:
            size = 0

        return size

    @override
    def delete(self, *args: Any, **kwargs: Any) -> tuple[int, dict[str, int]]:
        """Schedule deletion of the file in the store."""
        file_path = self.absolute_file_path()

        result = super().delete(*args, **kwargs)

        # If this method is called from a transaction: transaction.on_commit()
        # will call file_path.unlink when the most outer transaction is
        # committed.
        #
        # In the case that the code is running without a transaction:
        # the file_path.unlink will happen now.
        #
        # It's important that file_path.unlink is called only if the
        # DB is updated with the deletion. Otherwise, the file could be
        # deleted from the store but still referenced from the DB.
        transaction.on_commit(partial(file_path.unlink, missing_ok=True))

        return result

    def absolute_file_path(self) -> Path:
        """
        Return the absolute file path of the file.

        The files are stored in settings.DEBUSINE_UPLOAD_DIRECTORY.
        """
        return Path(settings.DEBUSINE_UPLOAD_DIRECTORY) / self.path


class ArtifactRelationQuerySet[A](QuerySet["ArtifactRelation", A]):
    """Custom QuerySet for ArtifactRelation."""

    def in_current_scope(self) -> "ArtifactRelationQuerySet[A]":
        """Filter to artifact relations in the current scope."""
        from debusine.db.context import context

        scope = context.require_scope()
        return self.filter(
            artifact__workspace__scope=scope, target__workspace__scope=scope
        )

    @permission_filter(work_request=Allow.PASS, anonymous=Allow.PASS)
    def can_display(
        self, pc: PermissionContext
    ) -> "ArtifactRelationQuerySet[A]":
        """Keep only ArtifactRelations that can be displayed."""
        # Delegate to workspace can_display check
        workspaces = Workspace.objects.can_display(pc)
        qs = self.filter(
            artifact__workspace__in=workspaces, target__workspace__in=workspaces
        )
        if pc.work_request is not None:
            task = pc.work_request.get_task()
            input_ids = task.get_input_artifacts_ids()
            qs = qs.filter(
                Q(artifact__pk__in=input_ids)
                | Q(artifact__created_by_work_request=pc.work_request)
            )
            qs = qs.filter(
                Q(target__pk__in=input_ids)
                | Q(target__created_by_work_request=pc.work_request)
            )
        return qs

    @permission_filter()
    def can_delete(
        self, pc: PermissionContext
    ) -> "ArtifactRelationQuerySet[A]":
        """Keep only ArtifactRelations that can be deleted."""
        # Delegate to workspace OWNERS
        owned_workspaces = Workspace.objects.with_role(
            pc, Workspace.Roles.OWNER
        )
        visible_workspaces = Workspace.objects.can_display(pc)
        return self.filter(
            artifact__workspace__in=owned_workspaces,
            target__workspace__in=visible_workspaces,
        )


class ArtifactRelationManager(models.Manager["ArtifactRelation"]):
    """Manager for the ArtifactRelation model."""

    @override
    def get_queryset(self) -> ArtifactRelationQuerySet[Any]:
        """Use the custom QuerySet."""
        return ArtifactRelationQuerySet(self.model, using=self._db)


class ArtifactRelation(models.Model):
    """Model relations between artifacts."""

    class Relations(models.TextChoices):
        EXTENDS = "extends", "Extends"
        RELATES_TO = "relates-to", "Relates to"
        BUILT_USING = "built-using", "Built using"

    artifact = models.ForeignKey(
        Artifact, on_delete=models.PROTECT, related_name="relations"
    )
    target = models.ForeignKey(
        Artifact, on_delete=models.PROTECT, related_name="targeted_by"
    )
    type = models.CharField(max_length=11, choices=Relations.choices)

    objects = ArtifactRelationManager.from_queryset(ArtifactRelationQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"
        constraints = [
            UniqueConstraint(
                fields=["artifact", "target", "type"],
                name="%(app_label)s_%(class)s_unique_artifact_target_type",
            )
        ]

    @override
    def __str__(self) -> str:
        """Return str for the object."""
        return f"{self.artifact.id} {self.type} {self.target.id}"

    @permission_check(
        "{user} cannot display artifact relation {resource}",
        work_request=Allow.PASS,
        anonymous=Allow.PASS,
    )
    def can_display(self, pc: PermissionContext) -> bool:
        """Check if the artifact can be displayed."""
        if not (
            self.artifact.workspace.can_display(pc)
            and self.target.workspace.can_display(pc)
        ):
            return False
        if pc.work_request is None:
            return True
        if (
            pc.work_request == self.artifact.created_by_work_request
            and pc.work_request == self.target.created_by_work_request
        ):
            return True
        task = pc.work_request.get_task()
        input_ids = task.get_input_artifacts_ids()
        return (
            pc.work_request == self.artifact.created_by_work_request
            or self.artifact.pk in input_ids
        ) and (
            pc.work_request == self.target.created_by_work_request
            or self.target.pk in input_ids
        )

    @permission_check(
        "{user} cannot delete artifact relation {resource}",
    )
    def can_delete(self, pc: PermissionContext) -> bool:
        """Check if the user can delete artifacts relations."""
        return self.artifact.workspace.has_role(
            pc, Workspace.Roles.OWNER
        ) and self.can_display(pc)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-client enrolling views."""

import asyncio
import json
from collections.abc import AsyncGenerator
from typing import Any

import django.db
from channels.layers import get_channel_layer
from django.db import transaction
from django.http import HttpRequest, HttpResponseBase, StreamingHttpResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View
from rest_framework import status

from debusine.client.models import EnrollPayload
from debusine.db.models.auth import ClientEnroll
from debusine.server.views import ProblemResponse


@method_decorator(transaction.non_atomic_requests, name="dispatch")
@method_decorator(csrf_exempt, name="dispatch")
class EnrollView(View):
    """Wait for a token after user confirmation."""

    #: Every this interval, send some data while waiting during the long poll
    PING_INTERVAL: int | float = 10

    def error_response(
        self,
        title: str,
        status_code: int = status.HTTP_500_INTERNAL_SERVER_ERROR,
    ) -> ProblemResponse:
        """Return an HTTP error."""
        return ProblemResponse(title=title, status_code=status_code)

    async def post(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        """Handle long-polling waiting for a token."""
        # Verify content-type
        if request.headers.get("Content-Type") != "application/json":
            return self.error_response(
                "only JSON data is accepted", status.HTTP_400_BAD_REQUEST
            )

        # Cap body length
        if len(request.body) > 4096:
            return self.error_response(
                "JSON payload too big", status.HTTP_400_BAD_REQUEST
            )

        # Decode the payload as JSON
        try:
            payload_unparsed = json.loads(request.body)
        except json.JSONDecodeError:
            return self.error_response(
                "payload is not valid JSON", status.HTTP_400_BAD_REQUEST
            )

        # Validate input with pydantic
        try:
            payload = EnrollPayload.model_validate(payload_unparsed)
        except ValueError:
            return self.error_response(
                "payload data is invalid", status.HTTP_400_BAD_REQUEST
            )

        # TODO: add info from http data? IP address, what else?

        # Store the request in the database
        try:
            self.object = await ClientEnroll.objects.acreate(
                nonce=payload.nonce, payload=payload.model_dump()
            )
        except django.db.DatabaseError:
            return self.error_response(
                "duplicate or invalid request received",
                status.HTTP_400_BAD_REQUEST,
            )

        return StreamingHttpResponse(
            self.long_poll(payload.nonce), content_type="application/json"
        )

    async def long_poll(self, nonce: str) -> AsyncGenerator[str]:
        """Handle long polling as a StreamingHttpResponse generator."""
        channel_layer = get_channel_layer()
        assert channel_layer is not None

        try:
            channel_name = f"enroll.{nonce}"
            confirm_task = asyncio.create_task(
                channel_layer.receive(channel_name)
            )
            while True:
                done, _ = await asyncio.wait(
                    [confirm_task], timeout=self.PING_INTERVAL
                )
                if not done:
                    # Generate some traffic to avoid nginx timing out the
                    # request. json.loads() will ignore leading whitespace,
                    # so we can use it for that purpose. See #857
                    yield "\n"
                else:
                    yield json.dumps(confirm_task.result())
                    break
        finally:
            await self.object.adelete()



# Generated by Django 4.2.28 on 2026-03-24 18:20

import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0049_token_worker_data'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='token_type',
            field=models.CharField(
                choices=[
                    ('worker-activation', 'Worker activation'),
                    ('worker', 'Worker'),
                    ('user', 'User'),
                ]
            ),
        ),
        migrations.AddConstraint(
            model_name='token',
            constraint=models.CheckConstraint(
                check=models.Q(
                    ('token_type__in', ('worker-activation', 'worker', 'user'))
                ),
                name='db_token_valid_type',
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='token',
            trigger=pgtrigger.compiler.Trigger(
                name='db_token_required_type_parameters',
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    constraint='CONSTRAINT',
                    func="SELECT * FROM db_token INTO NEW WHERE id = db_token.id; IF NEW.token_type = 'worker-activation' AND NEW.activating_worker_id IS NULL THEN RAISE EXCEPTION 'Worker activation token % has no activating worker', NEW.id; ELSIF NEW.token_type = 'worker' AND NEW.worker_id IS NULL THEN RAISE EXCEPTION 'Worker token % has no worker', NEW.id; ELSIF NEW.token_type = 'user' AND NEW.user_id IS NULL THEN RAISE EXCEPTION 'User token % has no user', NEW.id; ELSE RETURN NULL; END IF;",
                    hash='4b8d9585c5a2f20c3cfedadf18dab484ec219d56',
                    operation='INSERT OR UPDATE',
                    pgid='pgtrigger_db_token_required_type_parameters_a5603',
                    table='db_token',
                    timing='DEFERRABLE INITIALLY DEFERRED',
                    when='AFTER',
                ),
            ),
        ),
        migrations.AddConstraint(
            model_name='token',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        ('_worker_id__isnull', True),
                        ('token_type', 'worker-activation'),
                        ('user__isnull', True),
                    ),
                    models.Q(
                        ('_activating_worker_id__isnull', True),
                        ('token_type', 'worker'),
                        ('user__isnull', True),
                    ),
                    models.Q(
                        ('_activating_worker_id__isnull', True),
                        ('_worker_id__isnull', True),
                        ('token_type', 'user'),
                    ),
                    _connector='OR',
                ),
                name='db_token_consistent_type',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Server-side no-operation task."""

from typing import override

from debusine.artifacts.models import WorkRequestResults
from debusine.server.tasks import BaseServerTask
from debusine.server.tasks.models import ServerNoopData


class ServerNoop(BaseServerTask[ServerNoopData]):
    """Task that runs on server-side Celery workers and returns a boolean."""

    TASK_VERSION = 1

    @override
    def _execute(self) -> WorkRequestResults:
        """Act as specified by the client."""
        if self.data.exception:
            raise RuntimeError("Client requested an exception")
        return self.data.result

    @override
    def get_label(self) -> str:
        """Return the task label."""
        return "noop"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the DebianEnvironmentsManager."""

import datetime as dt
from typing import override

from django.utils import timezone

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.db.models import Collection, CollectionItem
from debusine.server.collections import (
    DebianEnvironmentsManager,
    ItemAdditionError,
)
from debusine.tasks.models import BackendType
from debusine.test.django import TestCase


class DebianEnvironmentsManagerTests(TestCase):
    """Tests for DebianEnvironmentsManager."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.user = self.playground.get_default_user()

        self.workspace = self.playground.get_default_workspace()

        self.collection = Collection.objects.create(
            name="Debian",
            category=CollectionCategory.ENVIRONMENTS,
            workspace=self.workspace,
        )

        self.manager = DebianEnvironmentsManager(collection=self.collection)

    def test_init_wrong_collection_category_raise_value_error(self) -> None:
        """Init raise ValueError: wrong collection category."""
        category = "debian:something-else"
        collection = Collection.objects.create(
            name="Name is not used",
            category=category,
            workspace=self.workspace,
        )

        msg = f'^DebianEnvironmentsManager cannot manage "{category}" category$'

        with self.assertRaisesRegex(ValueError, msg):
            DebianEnvironmentsManager(collection)

    def test_do_add_artifact(self) -> None:
        """Test do_add_artifact adds the artifact."""
        data = {"codename": "bookworm", "architecture": "amd64"}
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={**data, "with_dev": True},
        )

        collection_item = self.manager.add_artifact(artifact, user=self.user)

        collection_item.refresh_from_db()

        self.assertEqual(collection_item.name, "tarball:bookworm:amd64")
        self.assertEqual(
            collection_item.data, {**data, "variant": None, "backend": None}
        )

    def test_do_add_artifact_raise_item_addition_error(self) -> None:
        """Test do_add_artifact raise error: duplicated CollectionItem data."""
        data = {"codename": "bookworm", "architecture": "amd64"}
        artifact_1, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL, data=data
        )

        artifact_2, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=data,
        )

        self.manager.add_artifact(artifact_1, user=self.user)

        with self.assertRaisesRegex(
            ItemAdditionError, "db_collectionitem_unique_active_name"
        ):
            self.manager.add_artifact(artifact_2, user=self.user)

    def test_do_add_artifact_override_codename(self) -> None:
        """`do_add_artifact` can be told to override the codename."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={"codename": "bookworm", "architecture": "amd64"},
        )

        collection_item = self.manager.add_artifact(
            artifact, user=self.user, variables={"codename": "trixie"}
        )

        collection_item.refresh_from_db()

        self.assertEqual(collection_item.name, "tarball:trixie:amd64")
        self.assertEqual(
            collection_item.data,
            {
                "codename": "trixie",
                "architecture": "amd64",
                "variant": None,
                "backend": None,
            },
        )

    def test_do_add_artifact_variant(self) -> None:
        """`do_add_artifact` can be told to set a variant name."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_IMAGE,
            data={"codename": "bookworm", "architecture": "amd64"},
        )

        collection_item = self.manager.add_artifact(
            artifact, user=self.user, variables={"variant": "autopkgtest"}
        )

        collection_item.refresh_from_db()

        self.assertEqual(
            collection_item.name, "image:bookworm:amd64:autopkgtest:"
        )
        self.assertEqual(
            collection_item.data,
            {
                "codename": "bookworm",
                "architecture": "amd64",
                "variant": "autopkgtest",
                "backend": None,
            },
        )

    def test_do_add_artifact_backend(self) -> None:
        """`do_add_artifact` can be told to set a backend name."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_IMAGE,
            data={"codename": "bookworm", "architecture": "amd64"},
        )

        collection_item = self.manager.add_artifact(
            artifact, user=self.user, variables={"backend": BackendType.UNSHARE}
        )

        collection_item.refresh_from_db()

        self.assertEqual(collection_item.name, "image:bookworm:amd64::unshare")
        self.assertEqual(
            collection_item.data,
            {
                "codename": "bookworm",
                "architecture": "amd64",
                "variant": None,
                "backend": BackendType.UNSHARE,
            },
        )

    def test_do_add_artifact_replace(self) -> None:
        """do_add_artifact can replace an existing artifact."""
        workflow = self.playground.create_workflow()

        data = {"codename": "bookworm", "architecture": "amd64"}
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={**data, "with_dev": True},
        )
        collection_item = self.manager.add_artifact(artifact, user=self.user)
        artifact2, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={**data, "with_dev": True},
        )

        collection_item2 = self.manager.add_artifact(
            artifact2, user=self.user, replace=True, workflow=workflow
        )

        collection_item.refresh_from_db()
        self.assertEqual(collection_item.artifact, artifact)
        self.assertEqual(collection_item.removed_by_user, self.user)
        self.assertEqual(collection_item.removed_by_workflow, workflow)
        self.assertIsNotNone(collection_item.removed_at)
        self.assertEqual(collection_item2.name, "tarball:bookworm:amd64")
        self.assertEqual(collection_item2.artifact, artifact2)
        self.assertIsNone(collection_item2.removed_at)

    def test_do_add_artifact_replace_nonexistent(self) -> None:
        """Replacing a nonexistent artifact is allowed."""
        data = {"codename": "bookworm", "architecture": "amd64"}
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={**data, "with_dev": True},
        )

        collection_item = self.manager.add_artifact(
            artifact, user=self.user, replace=True
        )

        self.assertEqual(collection_item.name, "tarball:bookworm:amd64")
        self.assertEqual(collection_item.artifact, artifact)

    def test_do_add_collection_raise_item_addition_error(self) -> None:
        """
        Test do_add_collection raise ItemAdditionError.

        No Collections can be added or removed in
        debian:environments collection.
        """
        msg = (
            f'^Cannot add collections into '
            f'"{self.manager.COLLECTION_CATEGORY}"$'
        )
        collection = Collection.objects.create(
            name="Some-collection",
            category="Some category",
            workspace=self.workspace,
        )

        with self.assertRaisesRegex(ItemAdditionError, msg):
            self.manager.do_add_collection(collection, user=self.user)

    def test_do_remove_item_artifact(self) -> None:
        """``do_remove_item`` removes an artifact item."""
        data = {"codename": "bookworm", "architecture": "amd64"}

        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={**data, "with_dev": True},
        )

        collection_item = self.manager.add_artifact(artifact, user=self.user)

        # Test removing the artifact from the collection
        self.manager.remove_item(collection_item, user=self.user)

        # The artifact is not removed yet (retention period applies)
        self.assertEqual(collection_item.artifact, artifact)

        self.assertEqual(collection_item.removed_by_user, self.user)
        self.assertIsInstance(collection_item.removed_at, dt.datetime)

    def test_lookup_not_enough_colons_raise_lookup_error(self) -> None:
        """Test lookup raise LookupError: unexpected number of colons."""
        msg = '^Unexpected lookup format: "a"$'

        with self.assertRaisesRegex(LookupError, msg):
            self.manager.lookup("a")

    def test_lookup_unexpected_format_raise_lookup_error(self) -> None:
        """Test lookup raise LookupError: invalid format."""
        msg = '^Unexpected lookup format: "targz:codename=bookworm"$'

        with self.assertRaisesRegex(LookupError, msg):
            self.manager.lookup("targz:codename=bookworm")

    def test_lookup_return_none(self) -> None:
        """`lookup` returns None if there are no matches."""
        self.assertIsNone(
            self.manager.lookup("match:codename=bookworm:architecture=amd64")
        )
        self.assertIsNone(self.manager.lookup("name:nonexistent"))

    def test_lookup_return_matching_collection_item(self) -> None:
        """Test lookup return artifacts."""
        artifact_1, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={"codename": "bookworm", "architecture": "amd64"},
        )
        item_1 = self.manager.add_artifact(artifact_1, user=self.user)

        artifact_2, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={"codename": "bookworm", "architecture": "amd64"},
        )
        item_2 = self.manager.add_artifact(
            artifact_2,
            user=self.user,
            variables={"variant": "buildd", "backend": BackendType.UNSHARE},
        )

        artifact_3, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={"codename": "bookworm", "architecture": "i386"},
        )
        item_3 = self.manager.add_artifact(artifact_3, user=self.user)

        artifact_4, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={"codename": "trixie", "architecture": "amd64"},
        )
        item_4 = self.manager.add_artifact(artifact_4, user=self.user)

        artifact_5, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_IMAGE,
            data={"codename": "bookworm", "architecture": "amd64"},
        )
        item_5 = self.manager.add_artifact(artifact_5, user=self.user)

        # Next one is not returned by lookup because is removed.  (We
        # created it manually because self.manager.add_artifact would fail
        # with a constraint violation before we have a chance to mark the
        # artifact as being removed.)
        artifact_6, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_IMAGE,
            data={"codename": "bookworm", "architecture": "amd64"},
        )
        CollectionItem.objects.create(
            parent_collection=self.collection,
            name="image:bookworm:amd64",
            artifact=artifact_6,
            child_type=CollectionItem.Types.ARTIFACT,
            category=artifact_6.category,
            data={
                "codename": "bookworm",
                "architecture": "amd64",
                "variant": None,
                "backend": None,
            },
            created_by_user=self.user,
            removed_at=timezone.now(),
            removed_by_user=self.user,
        )

        # CollectionItem of type BARE should not exist in this collection
        # (the manager does not allow to add it). Add one to see that is
        # filtered out and not included in the result / cause problems
        CollectionItem.objects.create(
            child_type=CollectionItem.Types.BARE,
            created_by_user=self.user,
            parent_collection=self.collection,
            category="system:tarball",
            name="something",
            data={"codename": "bookworm", "arch": "amd64"},
        )

        # item_[1235] all match, but item_5 was created last.
        self.assertEqual(self.manager.lookup("match:codename=bookworm"), item_5)

        # item_[123] all match, but item_3 was created last.
        self.assertEqual(
            self.manager.lookup("match:format=tarball:codename=bookworm"),
            item_3,
        )

        # item_[124] all match, but item_4 was created last.
        self.assertEqual(
            self.manager.lookup("match:format=tarball:architecture=amd64"),
            item_4,
        )

        # item_1 and item_2 both match, but item_2 was created later.
        self.assertEqual(
            self.manager.lookup(
                "match:format=tarball:codename=bookworm:architecture=amd64"
            ),
            item_2,
        )

        self.assertEqual(
            self.manager.lookup(
                "match:format=tarball:codename=bookworm:architecture=amd64:"
                "variant="
            ),
            item_1,
        )
        self.assertEqual(
            self.manager.lookup(
                "match:format=tarball:codename=bookworm:architecture=amd64:"
                "variant=buildd"
            ),
            item_2,
        )
        self.assertEqual(
            self.manager.lookup(
                "match:format=tarball:codename=bookworm:architecture=amd64:"
                "backend=unshare"
            ),
            item_2,
        )
        self.assertEqual(
            self.manager.lookup(
                "match:format=image:codename=bookworm:architecture=amd64"
            ),
            item_5,
        )

        self.assertEqual(
            self.manager.lookup("name:tarball:bookworm:amd64"), item_1
        )
        self.assertEqual(
            self.manager.lookup("name:tarball:bookworm:amd64:buildd:unshare"),
            item_2,
        )
        self.assertEqual(
            self.manager.lookup("name:image:bookworm:amd64"), item_5
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Manager for debusine:task-history collections."""

import datetime as dt
import re
from typing import Any, TypedDict, override
from urllib.parse import quote, unquote

from django.db import IntegrityError
from django.db.models import IntegerField, Q, QuerySet, TextField, Value, Window
from django.db.models.fields.json import KT
from django.db.models.functions import Cast, Coalesce, Rank
from django.db.models.lookups import In
from django.utils import timezone

from debusine.artifacts.models import BareDataCategory, CollectionCategory
from debusine.db.models import CollectionItem, User, WorkRequest, Workspace
from debusine.db.permissioncontext import PermissionContext
from debusine.server.collections.base import (
    CollectionManagerInterface,
    ItemAdditionError,
)
from debusine.server.collections.lookup import lookup_multiple
from debusine.tasks.models import LookupMultiple, LookupSingle


class TaskHistoryManager(CollectionManagerInterface):
    """Manage collection of category debusine:task-history."""

    COLLECTION_CATEGORY = CollectionCategory.TASK_HISTORY
    VALID_BARE_DATA_CATEGORIES = frozenset(
        {BareDataCategory.HISTORICAL_TASK_RUN}
    )

    @override
    def do_add_bare_data(
        self,
        category: BareDataCategory,
        *,
        user: User,
        workflow: WorkRequest | None,
        data: dict[str, Any] | None = None,
        name: str | None = None,  # noqa: U100
        replace: bool = False,
        created_at: dt.datetime | None = None,
        replaced_by: CollectionItem | None = None,
    ) -> CollectionItem:
        """
        Add bare data into the managed collection.

        :param category: category of the bare data
        :param user: user adding the data to the collection
        :param workflow: workflow adding the data to the collection
        :param data: per-item data, used to compute the item name
        :param name: not used, must be None
        :param replace: if True, replace an existing similar item
        """
        if data is None:
            raise ItemAdditionError(
                f"Adding to {CollectionCategory.TASK_HISTORY} requires data"
            )
        if name is not None:
            raise ItemAdditionError(
                f"Cannot use an explicit item name when adding to "
                f"{CollectionCategory.TASK_HISTORY}"
            )

        work_request_id = data["work_request_id"]
        task_type = data["task_type"]
        task_name = data["task_name"]
        quoted_subject = quote(data.get("subject") or "")
        quoted_context = quote(data.get("context") or "")

        name_elements = [
            task_type,
            task_name,
            quoted_subject,
            quoted_context,
            str(work_request_id),
        ]
        name = ":".join(name_elements)

        if replace:
            self.remove_items_by_name(
                name=name,
                child_types=[CollectionItem.Types.BARE],
                user=user,
                workflow=workflow,
            )

        try:
            item = CollectionItem.objects.create_from_bare_data(
                category,
                parent_collection=self.collection,
                name=name,
                data=data,
                created_at=created_at,
                created_by_user=user,
                created_by_workflow=workflow,
                replaced_by=replaced_by,
            )
        except IntegrityError as exc:
            raise ItemAdditionError(str(exc))

        # For each subject/context combination, keep the last success, the last
        # failure, and a configured number of most recent entries.  Clean up
        # other matching items.
        old_items = self.collection.child_items.active().filter(
            child_type=CollectionItem.Types.BARE,
            category=category,
            data__task_type=task_type,
            data__task_name=task_name,
        )
        subject_context_rank = Window(
            Rank(),
            partition_by=(KT("data__subject"), KT("data__context")),
            order_by="-created_at",
        )
        last_success = (
            old_items.filter(data__result=WorkRequest.Results.SUCCESS)
            .annotate(subject_context_rank=subject_context_rank)
            .filter(subject_context_rank=1)
        )
        last_failure = (
            old_items.filter(
                data__result__in={
                    WorkRequest.Results.FAILURE,
                    WorkRequest.Results.ERROR,
                }
            )
            .annotate(subject_context_rank=subject_context_rank)
            .filter(subject_context_rank=1)
        )
        most_recent = old_items.annotate(
            subject_context_rank=subject_context_rank
        ).filter(
            subject_context_rank__lte=self.collection.data.get(
                "old_items_to_keep", 5
            )
        )
        for old_item in (
            old_items.exclude(id=item.id)
            .exclude(id__in=last_success.values("id"))
            .exclude(id__in=last_failure.values("id"))
            .exclude(id__in=most_recent.values("id"))
        ):
            self.remove_item(old_item, user=user, workflow=workflow)

        return item

    @override
    def do_remove_item(
        self,
        item: CollectionItem,
        *,
        user: User | None = None,
        workflow: WorkRequest | None = None,
    ) -> None:
        item.removed_by_user = user
        item.removed_by_workflow = workflow
        item.removed_at = timezone.now()
        item.save()

    @override
    def do_lookup(self, query: str) -> CollectionItem | None:
        """
        Return one CollectionItem based on the query.

        :param query: `last-entry:TASK_TYPE:TASK_NAME:SUBJECT:CONTEXT`,
          `last-success:TASK_TYPE:TASK_NAME:SUBJECT:CONTEXT`, or
          `last-failure:TASK_TYPE:TASK_NAME:SUBJECT:CONTEXT`.
        :raise LookupError: the query contains an unknown lookup format.
        """
        objects = self.collection.child_items.active()
        query_filter = Q(child_type=CollectionItem.Types.BARE)

        if m := re.match(
            r"^(last-entry|last-success|last-failure):(.+?):(.+?):(.*?):(.*)$",
            query,
        ):
            objects = objects.annotate(
                subject=Coalesce(
                    KT("data__subject"), Value("", output_field=TextField())
                ),
                context=Coalesce(
                    KT("data__context"), Value("", output_field=TextField())
                ),
            )
            query_filter &= Q(
                data__task_type=m.group(2),
                data__task_name=m.group(3),
                subject=unquote(m.group(4)),
                context=unquote(m.group(5)),
            )
            match m.group(1):
                case "last-entry":
                    pass
                case "last-success":
                    query_filter &= Q(data__result=WorkRequest.Results.SUCCESS)
                case "last-failure":
                    query_filter &= Q(
                        data__result__in={
                            WorkRequest.Results.FAILURE,
                            WorkRequest.Results.ERROR,
                        }
                    )
                case _ as unreachable:
                    raise AssertionError(
                        f'Unexpected lookup format: "{unreachable}" (expected '
                        f'"last-entry", "last-success", or "last-failure")'
                    )
            return objects.filter(query_filter).order_by("created_at").last()
        else:
            raise LookupError(
                f'Unexpected lookup format: "{query}" (expected '
                f'"last-entry", "last-success", or "last-failure")'
            )

    @override
    def do_lookup_filter(
        self,
        key: str,
        value: LookupSingle | LookupMultiple,
        *,
        workspace: Workspace,
        pc: PermissionContext,
        workflow_root: WorkRequest | None = None,
    ) -> Q:
        """
        Return :py:class:`CollectionItem` conditions for a lookup filter.

        :param key: For `same_work_request`, return conditions matching task
          runs for the same work request as any of the resulting artifacts.
          For `same_workflow`, return conditions matching task runs for work
          requests from the same workflow as any of the resulting artifacts.
        :raise LookupError: the query contains an unknown lookup filter
          format.
        """
        if key in {"same_work_request", "same_workflow"}:
            if isinstance(value, LookupSingle):
                value = LookupMultiple((value,))
            items = lookup_multiple(
                value, workspace, pc=pc, workflow_root=workflow_root
            )

            class IDValues(TypedDict):
                id: int

            work_request_ids: set[int] | QuerySet[WorkRequest, IDValues] = {
                item.artifact.created_by_work_request_id
                for item in items
                if (
                    item.artifact is not None
                    and item.artifact.created_by_work_request_id is not None
                )
            }
            if key == "same_workflow":
                work_request_ids = WorkRequest.objects.filter(
                    parent__children__in=work_request_ids
                ).values("id")
            return Q(
                In(
                    Cast("data__work_request_id", IntegerField()),
                    work_request_ids,
                )
            )
        else:
            raise LookupError(
                f'Unexpected lookup filter format: "{key}" (expected '
                f'"same_work_request" or "same_workflow")'
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""A custom Sphinx domain for Debusine-specific items."""

from collections.abc import Generator
from typing import Any, override

from docutils import nodes
from sphinx import addnodes
from sphinx.application import Sphinx
from sphinx.builders import Builder
from sphinx.directives import ObjectDescription
from sphinx.domains import Domain, ObjType
from sphinx.environment import BuildEnvironment
from sphinx.roles import XRefRole
from sphinx.util import logging
from sphinx.util.nodes import make_refnode

logger = logging.getLogger(__name__)


class BuildingBlockDirective(ObjectDescription[str]):
    """A custom directive describing a Debusine building block."""

    has_content = True
    required_arguments = 1

    block_type: str
    block_description: str

    @override
    def handle_signature(
        self, sig: str, signode: addnodes.desc_signature
    ) -> str:
        """Parse a signature and return a value that identifies the object."""
        signode += addnodes.desc_name(text=sig)
        return sig

    @override
    def add_target_and_index(
        self, name: str, sig: str, signode: addnodes.desc_signature
    ) -> None:
        """Add cross-reference IDs and entries to the index."""
        node_id = f"{self.block_type}-{name}"
        signode["ids"].append(node_id)
        self.indexnode["entries"].append(
            ("single", f"{name} ({self.block_description})", node_id, "", None)
        )
        domain = self.env.get_domain("debusine")
        assert isinstance(domain, DebusineDomain)
        domain.add_building_block(self.block_type, sig)


class ArtifactDirective(BuildingBlockDirective):
    """A custom directive describing an artifact."""

    block_type = "artifact"
    block_description = "artifact category"


class AssetDirective(BuildingBlockDirective):
    """A custom directive describing an asset."""

    block_type = "asset"
    block_description = "asset category"


class BareDataDirective(BuildingBlockDirective):
    """A custom directive describing bare data."""

    block_type = "bare-data"
    block_description = "bare data category"


class CollectionDirective(BuildingBlockDirective):
    """A custom directive describing a collection."""

    block_type = "collection"
    block_description = "collection category"


class FileBackendDirective(BuildingBlockDirective):
    """A custom directive describing a file backend."""

    block_type = "file-backend"
    block_description = "file backend"


class TaskDirective(BuildingBlockDirective):
    """A custom directive describing a task."""

    block_type = "task"
    block_description = "task name"


class WorkflowDirective(BuildingBlockDirective):
    """A custom directive describing a workflow."""

    block_type = "workflow"
    block_description = "workflow name"


class DebusineDomain(Domain):
    """A domain containing Debusine building blocks."""

    name = "debusine"
    label = "Debusine"
    object_types = {
        "artifact": ObjType("artifact", "artifact"),
        "asset": ObjType("asset", "asset"),
        "bare-data": ObjType("bare data", "bare-data"),
        "collection": ObjType("collection", "collection"),
        "file-backend": ObjType("file backend", "file-backend"),
        "task": ObjType("task", "task"),
        "workflow": ObjType("workflow", "workflow"),
    }
    roles = {
        "artifact": XRefRole(innernodeclass=nodes.inline),
        "asset": XRefRole(innernodeclass=nodes.inline),
        "bare-data": XRefRole(innernodeclass=nodes.inline),
        "collection": XRefRole(innernodeclass=nodes.inline),
        "file-backend": XRefRole(innernodeclass=nodes.inline),
        "task": XRefRole(innernodeclass=nodes.inline),
        "workflow": XRefRole(innernodeclass=nodes.inline),
    }
    directives = {
        "artifact": ArtifactDirective,
        "asset": AssetDirective,
        "bare-data": BareDataDirective,
        "collection": CollectionDirective,
        "file-backend": FileBackendDirective,
        "task": TaskDirective,
        "workflow": WorkflowDirective,
    }
    initial_data = {
        "objects": {},
    }

    @override
    def get_objects(self) -> Generator[tuple[str, str, str, str, str, int]]:
        """Return an iterable of object descriptions."""
        for (typ, name), (docname, node_id) in self.data["objects"].items():
            yield f"{typ}.{name}", name, typ, docname, node_id, 1

    @override
    def resolve_xref(
        self,
        env: BuildEnvironment,  # noqa: ARG002, U100
        fromdocname: str,
        builder: Builder,
        typ: str,
        target: str,
        node: addnodes.pending_xref,
        contnode: nodes.Element,
    ) -> nodes.reference | None:
        """Resolve a pending cross-reference with the given type and target."""
        objtypes = self.objtypes_for_role(typ)
        if not objtypes:
            return None
        for objtype in objtypes:
            result = self.data["objects"].get((objtype, target))
            if result:
                todocname, node_id = result
                return make_refnode(
                    builder,
                    fromdocname,
                    todocname,
                    node_id,
                    contnode,
                    f"{target} {objtype}",
                )
        logger.warning(
            "undefined %s: %r",
            typ,
            target,
            location=node,
            type="ref",
            subtype=node["reftype"],
        )
        return None

    def add_building_block(self, block_type: str, signature: str) -> None:
        """Add a new building block to the domain."""
        self.data["objects"][(block_type, signature)] = (
            self.env.docname,
            f"{block_type}-{signature}",
        )


def setup(app: Sphinx) -> dict[str, Any]:
    """Set up the extension."""
    app.add_domain(DebusineDomain)

    return {
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Utilities for emitting metrics."""

from collections.abc import Iterable
from logging import getLogger
from typing import Literal, Protocol, assert_never, override

from celery import current_task
from prometheus_client import Counter, Gauge, Histogram, Metric, Summary
from prometheus_client.registry import REGISTRY

from debusine.project.celery import get_celery_token

logger = getLogger(__name__)


class Collector(Protocol):
    """An object that supports metric collection."""

    def collect(self) -> Iterable[Metric]:
        """Collect metrics."""


class MetricNotFound(Exception):
    """The named metric was not found."""

    def __init__(self, name: str) -> None:
        """Construct the exception."""
        self.name = name

    @override
    def __str__(self) -> str:
        """Render the exception as a string."""
        return f"Metric {self.name!r} not found"


class BadMetricType(Exception):
    """The named metric has an unexpected type."""

    def __init__(self, name: str, metric_class_name: str) -> None:
        """Construct the exception."""
        self.name = name
        self.metric_class_name = metric_class_name

    @override
    def __str__(self) -> str:
        """Render the exception as a string."""
        return f"Metric {self.name!r} is not a {self.metric_class_name}"


def _check_metric_type[C: Collector](
    name: str, metric: Collector, metric_class: type[C]
) -> C:
    if not isinstance(metric, metric_class):
        raise BadMetricType(name, metric_class.__name__)
    return metric


def emit_metric(
    *,
    metric_type: Literal["counter", "gauge", "summary", "histogram"],
    name: str,
    labels: dict[str, str],
    value: float,
) -> None:
    """
    Emit a metric.

    If we're running in a Celery task, ask the server to emit the metric.
    Otherwise, do so directly.
    """
    if current_task and (celery_token := get_celery_token()) is not None:
        try:
            celery_token.client.emit_metric(
                metric_type=metric_type, name=name, labels=labels, value=value
            )
        except Exception as e:
            logger.warning(  # noqa: G200
                "Failed to report metric back to the server: %s", e
            )
    else:
        metric = REGISTRY._names_to_collectors.get(name)
        if metric is None:
            raise MetricNotFound(name)

        match metric_type:
            case "counter":
                _check_metric_type(name, metric, Counter).labels(**labels).inc(
                    value
                )
            case "gauge":
                _check_metric_type(name, metric, Gauge).labels(**labels).inc(
                    value
                )
            case "summary":
                _check_metric_type(name, metric, Summary).labels(
                    **labels
                ).observe(value)
            case "histogram":
                _check_metric_type(name, metric, Histogram).labels(
                    **labels
                ).observe(value)
            case _ as unreachable:
                assert_never(unreachable)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common code for workflows that handle regression tracking."""

import datetime as dt
from abc import ABCMeta
from functools import cached_property
from typing import TypeAlias, override

from django.db.models import Max, Window
from django.db.models.fields.json import KT
from django.db.models.functions import Greatest, Rank

from debusine.artifacts.models import ArtifactCategory, TaskTypes
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact, Collection, CollectionItem, WorkRequest
from debusine.server.workflows import workflow_utils
from debusine.server.workflows.base import Workflow, WorkflowValidationError
from debusine.server.workflows.models import RegressionTrackingWorkflowData
from debusine.tasks.models import (
    ActionSkipIfLookupResultChanged,
    BaseDynamicTaskData,
    RegressionAnalysisStatus,
)


class RegressionTrackingWorkflow[
    WD: RegressionTrackingWorkflowData,
    DTD: BaseDynamicTaskData,
](Workflow[WD, DTD], metaclass=ABCMeta):
    """Common code for a workflow that handles regression tracking."""

    @cached_property
    def qa_suite(self) -> Collection | None:
        """Suite that we are tracking regressions against, if configured."""
        if self.data.qa_suite is None:
            return None
        return self.work_request.lookup_single(
            self.data.qa_suite, expect_type=LookupChildType.COLLECTION
        ).collection

    @cached_property
    def qa_suite_changed(self) -> dt.datetime | None:
        """The latest time the ``qa_suite`` collection was changed."""
        # This method is only called when update_qa_results is "yes" or
        # "force", in which case this is checked by a model validator.
        assert self.qa_suite is not None

        changed = self.qa_suite.child_items.aggregate(
            changed=Greatest(Max("created_at"), Max("removed_at"))
        )["changed"]
        assert isinstance(changed, (dt.datetime, type(None)))
        return changed

    @cached_property
    def reference_qa_results(self) -> Collection | None:
        """Collection of reference results of QA tasks, if configured."""
        if self.data.reference_qa_results is None:
            return None
        return self.work_request.lookup_single(
            self.data.reference_qa_results,
            expect_type=LookupChildType.COLLECTION,
        ).collection

    @override
    def validate_input(self) -> None:
        """Thorough validation of input data."""
        # Validate that we can look up self.data.qa_suite and
        # self.data.reference_qa_results.
        try:
            self.qa_suite
            self.reference_qa_results
        except LookupError as e:
            raise WorkflowValidationError(str(e)) from e

    def skip_if_qa_result_changed(
        self,
        work_request: WorkRequest,
        *,
        package: str,
        architecture: str,
        promise_name: str | None = None,
    ) -> None:
        """Skip this work request if another workflow won the race."""
        lookup = (
            f"{self.data.reference_qa_results}/"
            f"latest:{work_request.task_name}_{package}_{architecture}"
        )
        try:
            item = self.work_request.lookup_single(lookup).collection_item
        except KeyError:
            item = None
        work_request.add_event_reaction(
            "on_assignment",
            ActionSkipIfLookupResultChanged(
                lookup=lookup,
                collection_item_id=None if item is None else item.id,
                promise_name=promise_name,
            ),
        )

    def find_reference_artifacts(self, task_name: str) -> dict[str, Artifact]:
        """Find the latest reference result for each architecture."""
        # This should be called from a workflow callback that is only
        # created when enable_regression_tracking is true, in which case
        # this is checked by a model validator.
        assert self.reference_qa_results is not None

        source_data = workflow_utils.source_package_data(self)
        reference_artifacts: dict[str, Artifact] = {}
        for result in (
            self.reference_qa_results.child_items.filter(
                child_type=CollectionItem.Types.ARTIFACT,
                artifact__isnull=False,
                data__task_name=task_name,
                data__package=source_data.name,
                data__has_key="architecture",
            )
            .annotate(
                rank_by_architecture=Window(
                    Rank(),
                    partition_by=KT("data__architecture"),
                    order_by="-data__timestamp",
                )
            )
            .filter(rank_by_architecture=1)
            .select_related("artifact__created_by_work_request")
        ):
            assert result.artifact is not None
            reference_artifacts[result.data["architecture"]] = result.artifact
        return reference_artifacts

    def find_new_artifacts(
        self, task_name: str, category: ArtifactCategory
    ) -> dict[str, Artifact]:
        """Find the new result for each architecture."""
        return {
            artifact.data["architecture"]: artifact
            for artifact in Artifact.objects.filter(
                created_by_work_request__in=(
                    self.work_request.children.unsuperseded()
                    .terminated()
                    .filter(task_type=TaskTypes.WORKER, task_name=task_name)
                ),
                category=category,
                data__has_key="architecture",
            ).select_related("created_by_work_request")
        }

    @staticmethod
    def compare_qa_results(
        reference: WorkRequest | None, new: WorkRequest | None
    ) -> RegressionAnalysisStatus:
        """
        Compare work requests providing two QA results.

        Return a status depending on the two work requests:

        * ``no-result``: when the comparison has not been completed yet
          (usually because we lack one of the two required QA results)
        * ``error``: when the comparison (or one of the required QA tasks)
          errored out
        * ``improvement``: when the new QA result is better than the
          original QA result
        * ``stable``: when the new QA result is neither better nor worse
          than the original QA result
        * ``regression``: when the new QA result is worse than the original
          QA result

        If the status is ``stable``, then the caller may perform a
        finer-grained analysis.
        """
        # Some aliases to make the match statement more readable.
        WR: TypeAlias = WorkRequest
        S: TypeAlias = WorkRequest.Statuses
        R: TypeAlias = WorkRequest.Results

        match (reference, new):
            case (None, _) | (_, None):
                return RegressionAnalysisStatus.NO_RESULT
            case WR(status=S.COMPLETED, result=R.SUCCESS | R.SKIPPED), WR(
                status=S.COMPLETED, result=R.FAILURE | R.ERROR
            ):
                return RegressionAnalysisStatus.REGRESSION
            case WR(status=S.COMPLETED, result=R.FAILURE | R.ERROR), WR(
                status=S.COMPLETED, result=R.SUCCESS | R.SKIPPED
            ):
                return RegressionAnalysisStatus.IMPROVEMENT
            case WR(status=S.COMPLETED, result=R.SUCCESS | R.SKIPPED), WR(
                status=S.COMPLETED, result=R.SUCCESS | R.SKIPPED
            ):
                return RegressionAnalysisStatus.STABLE
            case WR(status=S.COMPLETED, result=reference_result), WR(
                status=S.COMPLETED, result=new_result
            ) if reference_result == new_result:
                return RegressionAnalysisStatus.STABLE
            case _:
                return RegressionAnalysisStatus.ERROR



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Collection managers.

Provides interfaces to interact with Collections: add items, remove,
lookups, etc.

Each specific CollectionManager implement their own business logic.
"""

from django.conf import settings

from debusine.server.collections.base import (
    CollectionManagerInterface,
    ItemAdditionError,
    ItemRemovalError,
)
from debusine.server.collections.debian_archive import DebianArchiveManager
from debusine.server.collections.debian_environments import (
    DebianEnvironmentsManager,
)
from debusine.server.collections.debian_package_build_logs import (
    DebianPackageBuildLogsManager,
)
from debusine.server.collections.debian_qa_results import DebianQAResultsManager
from debusine.server.collections.debian_suite import DebianSuiteManager
from debusine.server.collections.debusine_task_configuration import (
    DebusineTaskConfigurationManager,
)
from debusine.server.collections.task_history import TaskHistoryManager
from debusine.server.collections.workflow_internal import (
    WorkflowInternalManager,
)

__all__ = [
    "CollectionManagerInterface",
    "DebianArchiveManager",
    "DebianEnvironmentsManager",
    "DebianPackageBuildLogsManager",
    "DebianQAResultsManager",
    "DebianSuiteManager",
    "DebusineTaskConfigurationManager",
    "ItemAdditionError",
    "ItemRemovalError",
    "TaskHistoryManager",
    "WorkflowInternalManager",
]

if getattr(settings, "TEST_MODE", False):  # pragma: no cover
    # Register a manager to make debusine:test collections easily usable by
    # any test.  This manager doesn't implement any specific semantics, but
    # it at least allows basic name lookups to work.
    from debusine.server.collections.tests.test_base import (  # noqa: F401
        DebusineTestManager,
    )

    __all__.append("DebusineTestManager")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for tabular representation of querysets."""

import abc
from typing import override
from urllib.parse import urlencode

from django.core.exceptions import ImproperlyConfigured
from django.test import RequestFactory

from debusine.db.models import User, Workspace
from debusine.test.django import TestCase
from debusine.web.views.table import Column, FilterText, Ordering, Table
from debusine.web.views.tests.utils import ViewTestMixin


class TableTestCase(ViewTestMixin, TestCase, abc.ABC):
    """Common test infrastructure."""

    @abc.abstractmethod
    def get_table_class(self) -> type[Table[User]]:
        """Get the table class to use for tests."""

    def _table(
        self,
        prefix: str = "",
        default_order: str | None = None,
        table_class: type[Table[User]] | None = None,
        qstring: dict[str, str] | None = None,
        preview: bool = False,
        fragment: str | None = None,
    ) -> Table[User]:
        """Instantiate a table."""
        url = "/"
        if qstring:
            url += "?" + urlencode(qstring)
        request = RequestFactory().get(url)
        queryset = User.objects.all()
        if table_class is None:
            table_class = self.get_table_class()
        return table_class(
            request=request,
            object_list=queryset,
            prefix=prefix,
            default_order=default_order,
            preview=preview,
            fragment=fragment,
        )


class TableTests(TableTestCase):
    """Tests for :py:class:`Table`."""

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        for idx in range(10):
            cls.playground.create_user(f"user{idx:02d}")

    @override
    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            email = Column("Email", ordering="email")
            test = Column("Test")
            default_order = "user"

        return _Table

    def test_declarative(self) -> None:
        """Test declarative column definition."""

        class _Table(Table[User]):
            user = Column("User", ordering="username")
            default_order = "user"

        self.assertEqual(list(_Table.column_definitions.keys()), ["user"])

        table = self._table(table_class=_Table)
        self.assertIs(table.model, User)
        self.assertEqual(list(table.columns.columns.keys()), ["user"])
        self.assertEqual(
            [f"{o.what}.{o.how}" for o in table.orderings.options],
            ["user.asc", "user.desc"],
        )
        self.assertEqual(list(table.filters.options.keys()), [])

        col = table.columns["user"]
        self.assertEqual(col.orderings, table.orderings.options)

    def test_declarative_aliased(self) -> None:
        """Test declarative column definition with an aliased name."""

        class _Table(Table[User]):
            type_ = Column("Type", ordering="username", name="type")
            default_order = "type"

        self.assertEqual(list(_Table.column_definitions.keys()), ["type"])

        table = self._table(table_class=_Table)
        self.assertEqual(list(table.columns.columns.keys()), ["type"])

        self.assertEqual(
            [f"{o.what}.{o.how}" for o in table.orderings.options],
            ["type.asc", "type.desc"],
        )
        self.assertEqual(list(table.filters.options.keys()), [])

        col = table.columns["type"]
        self.assertEqual(col.orderings, table.orderings.options)

    def test_declarative_inherited(self) -> None:
        """Test declarative column inherited from base classes."""

        class _Base(Table[User]):
            name = Column("Name")

        class _Derived(_Base):
            email = Column("Email")

        self.assertEqual(list(_Base.column_definitions), ["name"])
        self.assertEqual(list(_Derived.column_definitions), ["name", "email"])

    def test_declarative_non_table_base(self) -> None:
        """Test declarative column inherited from base classes."""

        class Mixin:
            pass

        class _Base(Table[User]):
            name = Column("Name")

        class _Derived(Mixin, _Base):
            email = Column("Email")

        self.assertEqual(list(_Base.column_definitions), ["name"])
        self.assertEqual(list(_Derived.column_definitions), ["name", "email"])

    def test_declarative_multiple_inheritance(self) -> None:
        """Test declarative column inherited from base classes."""

        class _Base1(Table[User]):
            name = Column("Name")

        class _Base2(Table[User]):
            status = Column("Status")

        class _Derived(_Base1, _Base2):
            email = Column("Email")

        self.assertEqual(list(_Base1.column_definitions), ["name"])
        self.assertEqual(list(_Base2.column_definitions), ["status"])
        self.assertEqual(
            list(_Derived.column_definitions), ["name", "status", "email"]
        )

    def test_declarative_filter(self) -> None:
        """Test declarative table filters."""

        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_username = FilterText("User")
            default_order = "user"

        self.assertEqual(list(_Table.filter_definitions.keys()), ["username"])

        table = self._table(table_class=_Table)
        self.assertEqual(list(table.filters.options.keys()), ["username"])
        flt = table.filters.options["username"]
        self.assertEqual(flt.name, "username")
        self.assertEqual(flt.label, "User")

    def test_declarative_filter_prefix(self) -> None:
        """Declarative filter members must start with ``filter_``."""
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"f_username = FilterText\(…\): assigned name"
            r" does not start with 'filter_'",
        ):

            class _Table(Table[User]):
                user = Column("User", ordering="username")
                f_username = FilterText("User")
                default_order = "user"

    def test_declarative_ordering(self) -> None:
        """Test declarative table orderings."""

        class _Table(Table[User]):
            user = Column("User")
            ordering_test = Ordering("username")
            default_order = "test"

        self.assertEqual(list(_Table.ordering_definitions.keys()), ["test"])

        table = self._table(table_class=_Table)
        self.assertEqual(
            [str(o) for o in table.orderings.options], ["test.asc", "test.desc"]
        )
        ordering = table.orderings.options[0]
        self.assertEqual(ordering.what, "test")
        self.assertEqual(ordering.how, "asc")
        self.assertEqual(ordering.label_what, "Test")

    def test_mandate_queryset(self) -> None:
        """Table needs a QuerySet for object_list."""
        table_class = self.get_table_class()
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"Table cannot work with an object_list of class list,"
            r" only with QuerySets",
        ):
            table_class(
                request=RequestFactory().get("/"),
                object_list=list(User.objects.all()),
            )

    def test_missing_default_order(self) -> None:
        """Test with no default ordering."""

        class _Table(Table[User]):
            """Table with default ordering not set."""

            user = Column("User", ordering="username")

        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"default_order not set in class or constructor arguments",
        ):
            self._table(table_class=_Table)

    def test_unsortable(self) -> None:
        """Test a table with no ordering options."""

        class _Table(Table[User]):
            """Unsortable table."""

            user = Column("User")

        self.assertEqual(list(_Table.column_definitions.keys()), ["user"])
        self.assertEqual(_Table.filter_definitions, {})

        table = self._table(table_class=_Table)
        self.assertIs(table.model, User)
        self.assertEqual(list(table.columns.columns.keys()), ["user"])
        self.assertEqual(table.orderings.options, [])
        self.assertEqual(table.filters.options, {})

        col = table.columns["user"]
        self.assertEqual(col.orderings, [])

        self.assertQuerySetEqual(
            table.rows, list(User.objects.all()), ordered=False
        )

    def test_preview_disables_filters(self) -> None:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            filter_username = FilterText("User")
            default_order = "user"

        table = self._table(table_class=_Table, preview=True)
        self.assertEqual(list(table.filters.options.keys()), [])

    def test_make_link(self) -> None:
        table = self._table()
        for kwargs, expected in (
            ({}, ""),
            ({"test": 1}, "test=1"),
        ):
            with self.subTest(kwargs=kwargs):
                self.assertEqual(table.make_link(**kwargs), expected)

    def test_make_link_fragment(self) -> None:
        table = self._table(fragment="foo")
        for kwargs, expected in (
            ({}, "#foo"),
            ({"test": 1}, "test=1#foo"),
        ):
            with self.subTest(kwargs=kwargs):
                self.assertEqual(table.make_link(**kwargs), expected)

    def test_related_fields(self) -> None:
        class _Table(Table[Workspace]):
            workspace = Column("Workspace", ordering="name")
            default_order = "workspace"
            related_fields = ("scope",)

        for i in range(3):
            self.playground.create_workspace(name=f"test-{i}")
        table = _Table(
            request=RequestFactory().get("/"),
            object_list=Workspace.objects.all(),
        )
        expected_workspaces = list(Workspace.objects.order_by("name"))
        with self.assertNumQueries(1):
            rows = table.rows
            self.assertQuerySetEqual(rows, expected_workspaces)
        with self.assertNumQueries(0):
            [row.scope.id for row in rows]



# Generated by Django 4.2.20 on 2025-05-06 16:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0005_collectionitem_db_ci_suite_index_path_idx'),
    ]
    replaces = [
        ('db', '0147_fileinartifact_content_type'),
    ]

    operations = [
        migrations.AddField(
            model_name='fileinartifact',
            name='content_type',
            field=models.CharField(blank=True, null=True),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Receive server notifications."""

import asyncio
import logging
import types
from collections.abc import AsyncGenerator
from typing import Any, NoReturn, Self

import aiohttp
import tenacity

from debusine.client.debusine import Debusine, ServerNotificationDecodeError
from debusine.client.exceptions import TokenDisabledError
from debusine.utils import get_user_agent

logger = logging.getLogger(__name__)


class ServerNotifications:
    """Read asynchronous notification from a remote Debusine server."""

    def __init__(self, client: Debusine, endpoint: str) -> None:
        """
        Initialize ServerNotifications.

        :param config: worker configuration
        :param client: scope-less Debusine client
        """
        self.client = client
        if self.client.token is None:
            raise ValueError(
                "Debusine client needs to be configured with an API token"
            )
        self.endpoint = endpoint
        self.message_queue: asyncio.Queue[dict[str, Any]] = asyncio.Queue()
        self.client_session: aiohttp.ClientSession | None = None
        self.debusine_ws_url = self.client.base_api_url.replace('http', 'ws', 1)
        self.worker_connect_ws_url = f"{self.debusine_ws_url}/ws/{endpoint}"
        # Set to True if the retry_attempt_number should be reset
        self._reset_retry_attempt_number: bool = False

        # Running main loop
        self.main_loop_task: asyncio.Task[None] | None = None

        # Set to True to stop the main loop from looping
        self._shutting_down: bool = False

    @staticmethod
    async def _asyncio_sleep(delay: float) -> None:
        """Sleep asynchronously (mocked in tests)."""
        await asyncio.sleep(delay)

    def _tenacity_before_hook(
        self, retry_state: tenacity.RetryCallState
    ) -> None:
        self._reset_retry_counter(retry_state)
        self._log_trying_to_reconnect(retry_state)

    def _reset_retry_counter(
        self, retry_state: tenacity.RetryCallState
    ) -> None:
        if self._reset_retry_attempt_number:
            retry_state.attempt_number = 0
            self._reset_retry_attempt_number = False

    @staticmethod
    def _log_trying_to_reconnect(retry_state: tenacity.RetryCallState) -> None:
        if retry_state.attempt_number > 1:
            logger.debug(
                'Trying to reconnect, attempt %d', retry_state.attempt_number
            )

    def _raise_try_again(self) -> NoReturn:
        raise tenacity.TryAgain()

    async def open(self) -> None:
        """Start the reader."""
        if self.main_loop_task is not None:
            raise AssertionError("Reader task already started")
        self.main_loop_task = asyncio.create_task(self.main_loop())

    async def close(self) -> None:
        """Close the message reader."""
        self._shutting_down = True
        if self.client_session is not None:
            await self.client_session.close()

        if self.main_loop_task is not None:
            self.main_loop_task.cancel()
            # Await so that we do not hide potential exceptions raised in
            # main_loop
            try:
                await self.main_loop_task
            except asyncio.CancelledError:
                pass
            self.main_loop_task = None

    async def __aenter__(self) -> Self:
        """Start the reader task."""
        await self.open()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        """Call ``close()`` on exit."""
        await self.close()

    async def messages(self) -> AsyncGenerator[dict[str, Any]]:
        """Generate messages received from the server."""
        if self.main_loop_task is None:
            raise AssertionError("Main loop task not started")

        # If close() is called during iteration self.main_loop_task can become
        # None, so we need to handle it here
        while self.main_loop_task and not self.main_loop_task.done():
            get_task = asyncio.create_task(self.message_queue.get())

            done, _ = await asyncio.wait(
                (self.main_loop_task, get_task),
                return_when=asyncio.FIRST_COMPLETED,
            )
            if get_task in done:
                payload = get_task.result()
                yield payload
                self.message_queue.task_done()

    async def main_loop(self) -> None:
        """
        Read notifications from the server and add them to ``message_queue``.

        Retries and reconnects are transparently handled.
        """
        reader = tenacity.retry(
            sleep=self._asyncio_sleep,
            wait=tenacity.wait_random(min=1, max=6),
            before=self._tenacity_before_hook,
            retry=tenacity.retry_if_exception_type(
                aiohttp.client_exceptions.ClientError
            )
            | tenacity.retry_if_exception_type(TokenDisabledError)
            | tenacity.retry_if_exception_type(ConnectionError),
        )(self._wait_for_messages)
        await reader()
        logger.debug("Server notification reading loop ended")

    async def _wait_for_messages(self) -> None:
        """Connect to the server and waits for commands."""
        # Enforced by constructor
        assert self.client.token is not None
        headers = {'token': self.client.token}
        async with aiohttp.ClientSession(
            headers={"User-Agent": get_user_agent()}
        ) as client_session:
            self.client_session = client_session

            try:
                async with self.client_session.ws_connect(
                    self.worker_connect_ws_url, headers=headers, heartbeat=60
                ) as ws:
                    logger.info("Websocket connection has been established")
                    self._reset_retry_attempt_number = True
                    msg: aiohttp.WSMessage
                    async for msg in ws:  # pragma: no branch
                        logger.debug(
                            "Incoming websocket message (%s): %s",
                            msg.type.name,
                            str(msg.data),
                        )
                        try:
                            payload = self.client.decode_ws_message(msg)
                        except ServerNotificationDecodeError as exc:
                            logger.warning(
                                "Unhandled websocket notification: %s", exc
                            )
                            continue

                        await self.message_queue.put(payload)

                    logger.info("Server notification channel closed")

                    if not self._shutting_down:
                        # The server disconnected the worker or we lost the
                        # connection
                        logger.warning(
                            "Websocket connection got closed (close_code=%s)",
                            ws.close_code,
                        )
                        self._raise_try_again()

            except TokenDisabledError as exc:
                # The server disconnected the worker because the token is
                # not associated with a worker or is disabled. The server
                # admin should enable the token.
                logger.info(  # noqa: G200
                    'The token is disabled. '
                    'Debusine admin (%s) should enable it. '
                    'Reason: %s. '
                    'Will try again',
                    self.debusine_ws_url,
                    exc,
                )
                raise
            except (
                aiohttp.client_exceptions.ClientConnectorError,
                aiohttp.client_exceptions.WSServerHandshakeError,
            ) as exc:
                logger.info(  # noqa: G200
                    'Error connecting to %s (%s)',
                    self.debusine_ws_url,
                    exc,
                )
                raise
            except ConnectionResetError:
                logger.debug('ConnectionResetError, will reconnect')
                raise

        self.client_session = None



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine task configuration repository management."""

import difflib
import logging
import shutil
import subprocess
from collections import defaultdict
from collections.abc import Generator, Iterable
from pathlib import Path
from typing import Any, Literal, Self, TypeAlias, assert_never, override

import yaml

from debusine.artifacts.models import DebusineTaskConfiguration
from debusine.client.models import StrictBaseModel, TaskConfigurationCollection


class Item(DebusineTaskConfiguration):
    """DebusineTaskConfiguration with path and position never set."""

    path: None = None
    position: None = None


def as_item(value: DebusineTaskConfiguration) -> Item:
    """Convert a DebusineTaskConfiguration to an Item."""
    serialized = value.model_dump()
    serialized.pop("path", None)
    serialized.pop("position", None)
    return Item.model_validate(serialized)


class InvalidRepository(Exception):
    """The local repository is not valid or does not match the remote one."""


class PullStats:
    """Statistics for a pull operation."""

    __slots__ = "added", "deleted", "unchanged", "updated"

    def __init__(
        self,
        added: int = 0,
        updated: int = 0,
        deleted: int = 0,
        unchanged: int = 0,
    ) -> None:
        """Initialize a new PullStats."""
        self.added = added
        self.updated = updated
        self.deleted = deleted
        self.unchanged = unchanged

    @override
    def __str__(self) -> str:
        return (
            f"{self.added} added, "
            f"{self.updated} updated, "
            f"{self.deleted} deleted, "
            f"{self.unchanged} unchanged"
        )

    @override
    def __repr__(self) -> str:
        return (
            f"PullStats(added={self.added}, updated={self.updated},"
            f" deleted={self.deleted}, unchanged={self.unchanged})"
        )

    def __iadd__(self, other: "PullStats") -> Self:
        """Add values from another PullStats."""
        self.added += other.added
        self.updated += other.updated
        self.deleted += other.deleted
        self.unchanged += other.unchanged
        return self

    @override
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, PullStats):
            return NotImplemented

        return (self.added, self.updated, self.deleted, self.unchanged) == (
            other.added,
            other.updated,
            other.deleted,
            other.unchanged,
        )

    def has_changes(self) -> bool:
        """Check if anything was changed."""
        return self.added > 0 or self.updated > 0 or self.deleted > 0


class Manifest(StrictBaseModel):
    """Repository metadata."""

    workspace: str
    collection: TaskConfigurationCollection


class EntryFile:
    """A file with task configuration entries."""

    relpath: Path
    entries: list[Item]

    def __init__(self, relpath: Path) -> None:
        """Create an empty EntryFile."""
        self.relpath = relpath
        self.entries = []

    def load(self, root: Path) -> None:
        """Load contents from the file system."""
        path = root / self.relpath
        try:
            with path.open() as fd:
                data = yaml.safe_load(fd)

            if isinstance(data, list):
                for item in data:
                    item.pop("path", None)
                    item.pop("position", None)
                    self.entries.append(Item.model_validate(item))
            else:
                data.pop("path", None)
                data.pop("position", None)
                self.entries.append(Item.model_validate(data))
        except Exception as e:
            e.add_note(f"source: {self.relpath}")
            raise

    def write(self, root: Path) -> None:
        """
        Write the contents to a file.

        If the file already exists and its contents would be unchanged, it is
        left untouched.
        """
        # Load the old contents
        old = EntryFile(self.relpath)
        try:
            old.load(root)
        except FileNotFoundError:
            pass
        else:
            # If they are unchanged, do nothing
            if self.entries == old.entries:
                return

        # Else rewrite the file
        data: list[dict[str, Any]] | dict[str, Any]
        if len(self.entries) == 1:
            data = self.entries[0].model_dump(mode="json")
        else:
            data = [e.model_dump(mode="json") for e in self.entries]
        path = root / self.relpath
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("wt") as fd:
            yaml.safe_dump(data, fd)

    def diff(self, other: "EntryFile") -> PullStats:
        """Compute change statistics."""
        matcher = difflib.SequenceMatcher(
            a=[e.model_dump_json() for e in self.entries],
            b=[e.model_dump_json() for e in other.entries],
            autojunk=False,
        )
        stats = PullStats()
        tag: Literal["replace", "delete", "insert", "equal"]
        for tag, i1, i2, j1, j2 in matcher.get_opcodes():
            match tag:
                case "replace":
                    # a[i1:i2] should be replaced by b[j1:j2]
                    ra = i2 - i1
                    rb = j2 - j1
                    if ra > rb:
                        stats.deleted += ra - rb
                        stats.updated += rb
                    elif ra < rb:
                        stats.added += rb - ra
                        stats.updated += ra
                    else:
                        stats.updated += ra
                case "delete":
                    # a[i1:i2] should be deleted
                    stats.deleted += i2 - i1
                case "insert":
                    # b[j1:j2] should be inserted at a[i1:i1]
                    stats.added += j2 - j1
                case "equal":
                    # a[i1:i2] == b[j1:j2] (the sub-sequences are equal)
                    stats.unchanged += i2 - i1
                case _ as unreachable:
                    assert_never(unreachable)

        # Note: updated is now representing items being replaced in the same
        # file position. We do not have the concept of an entry being updated
        # anymore based on its identifying information, because we now accept
        # multiple entries in a file that match the same task information
        return stats


class RemoteTaskConfigurationRepository:
    """Task configuration repository populated from a remote server."""

    #: Store an alias to the Manifest type
    Manifest: TypeAlias = Manifest

    def __init__(self, manifest: Manifest) -> None:
        """Initialize with a manifest."""
        #: Repository metadata
        self.manifest = manifest
        #: Task configuration entries
        self.entries: list[DebusineTaskConfiguration] = []

    @classmethod
    def from_items(
        cls, manifest: Manifest, items: Iterable[DebusineTaskConfiguration]
    ) -> Self:
        """Load contents from a stream of items."""
        res = cls(manifest)
        res.entries.extend(items)
        return res


class LocalTaskConfigurationRepository:
    """Task configuration repository stored in the local file system."""

    #: Store an alias to the Manifest type
    Manifest: TypeAlias = Manifest

    def __init__(self, root: Path) -> None:
        """Initialize an empty repository."""
        #: Repository metadata
        self.manifest: Manifest | None
        #: Task configuration entries
        self.files: dict[Path, EntryFile] = {}

        #: On-disk root
        self.root = root
        #: Path of the manifest file
        self.manifest_path = self.root / "MANIFEST"

        self.reload()

    def reload(self) -> None:
        """Load/reload contents."""
        self.manifest = None
        self.files = {}

        #: Load manifest from disk
        if self.manifest_path.exists():
            self.manifest = self.read_manifest(self.root)

        #: Load items from disk
        for relpath in self.iter_files(self.root):
            entry_file = EntryFile(relpath)
            entry_file.load(self.root)
            self.files[relpath] = entry_file

    def iter_entries(self) -> Generator[DebusineTaskConfiguration]:
        """
        Return the list of items in the repository.

        Items are annotated with file and position information.
        """
        for entry_file in self.files.values():
            for position, entry in enumerate(entry_file.entries):
                e = DebusineTaskConfiguration.model_validate(entry.model_dump())
                e.path = entry_file.relpath.as_posix()
                e.position = position
                yield e

    def is_dirty(self) -> bool:
        """Check if the git repository is dirty."""
        return False

    def git_commit(self) -> str | None:
        """Return the hash of the current git commit."""
        return None

    def has_commit(self, commit: str) -> bool:  # noqa: ARG002, U100
        """Check if the given commit hash is in our history."""
        return False

    def write_manifest(self) -> None:
        """Write repository manifest from disk."""
        if self.manifest is None:
            raise InvalidRepository("Repository manifest has not been set")
        try:
            with self.manifest_path.open("wt") as fd:
                yaml.safe_dump(self.manifest.model_dump(), fd)
        except Exception as exc:
            exc.add_note(f"manifest: {self.manifest_path}")
            raise

    def _merge_manifest(self, manifest: Manifest) -> None:
        if self.manifest is None:
            self.manifest = manifest.model_copy()
        elif (
            self.manifest.workspace != manifest.workspace
            or self.manifest.collection.id != manifest.collection.id
            or self.manifest.collection.name != manifest.collection.name
        ):
            raise InvalidRepository(
                "repo to pull refers to collection"
                f" {manifest.workspace}/{manifest.collection.name}"
                f" ({manifest.collection.id})"
                " while the checkout has collection"
                f" {self.manifest.workspace}/{self.manifest.collection.name}"
                f" ({self.manifest.collection.id})"
            )
        else:
            self.manifest.collection.data = manifest.collection.data.copy()

    def find_in_repo(
        self, entry: DebusineTaskConfiguration
    ) -> Generator[DebusineTaskConfiguration]:
        """
        Try to find the given entry in the repo.

        The entry is matched by exact contents.

        Generates all matching entries, annotated with path and position
        information.
        """
        item = as_item(entry)
        for entry_file in self.files.values():
            for idx, file_item in enumerate(entry_file.entries):
                if item == file_item:
                    found = entry.model_copy()
                    found.path = entry_file.relpath.as_posix()
                    found.position = idx
                    yield found

    def _pull(
        self,
        repo: RemoteTaskConfigurationRepository,
        logger: logging.Logger,
        preserve_deleted: bool = True,
    ) -> PullStats:
        """
        Pull changes from another repository.

        :param repo: repository to pull from
        :param logger: logger to use to give user feedback about what changed
        :param preserve_deleted: if True, items deleted in repo will be
               preserved in self; if False, items deleted in repo will be
               deleted in self.
        """
        # Divide the entries that have a path from those that do not
        pathless: list[DebusineTaskConfiguration] = []
        # And dispatch entries by path
        by_path: dict[Path, list[DebusineTaskConfiguration]] = defaultdict(list)
        for entry in repo.entries:
            # Try to fill path information by looking up the entry in the local
            # repository
            if entry.path is None:
                found = False
                for match in self.find_in_repo(entry):
                    found = True
                    assert match.path is not None
                    by_path[Path(match.path)].append(match)
                if not found:
                    # Ensure entries without path also do not have a position
                    entry.position = None
                    pathless.append(as_item(entry))
            else:
                by_path[Path(entry.path)].append(entry)

        # Dispatch pathless entries in new.yaml
        if pathless:
            new_file = by_path[Path("new.yaml")]
            new_file.extend(pathless)

        # Build EntryFile instances from dispatched entries
        new_files: dict[Path, EntryFile] = {}
        for relpath, entries in by_path.items():
            position_when_unset = max(e.position or 0 for e in entries) + 1
            entries.sort(
                key=lambda e: (
                    e.position
                    if e.position is not None
                    else position_when_unset
                )
            )

            entry_file = EntryFile(relpath)
            entry_file.entries.extend(as_item(e) for e in entries)
            new_files[relpath] = entry_file

        stats = PullStats()
        self._merge_manifest(repo.manifest)

        # Replace repository contents in memory
        old_files = self.files
        self.files = new_files

        # Compute and show differences
        for relpath in old_files.keys() & new_files.keys():
            old = old_files[relpath]
            new = new_files[relpath]
            file_stats = old.diff(new)
            stats += file_stats
            if file_stats.has_changes():
                logger.info("%s: %s", relpath, file_stats)

        for relpath in new_files.keys() - old_files.keys():
            new = new_files[relpath]
            logger.info(
                "%s: new file with %d entries", relpath, len(new.entries)
            )
            stats.added += len(new.entries)

        for relpath in list(old_files.keys() - new_files.keys()):
            old = old_files[relpath]
            action = (
                "preserved from deletion" if preserve_deleted else "deleted"
            )
            logger.info(
                "%s: file with %d entries %s", relpath, len(old.entries), action
            )
            stats.deleted += len(old.entries)
            if preserve_deleted:
                self.files[relpath] = old

        # Write out results to filesystem
        self.write_manifest()
        self.write_entries()

        return stats

    def clean_empty_dirs(self, deleted_paths: list[Path]) -> None:
        """Remove leftover empty directories after file deletions."""
        # Compute the set of all affected parent directories
        parents: set[Path] = set()
        for path in deleted_paths:
            parents.update(path.parents[:-1])

        # Sort by descending number of path components
        todo = sorted(parents, key=lambda p: len(p.parents), reverse=True)

        # Delete all parents if empty
        for path in todo:
            try:
                (self.root / path).rmdir()
            except OSError:
                pass

    def write_entries(self) -> None:
        """Write entries to disk."""
        # List files that need to be deleted
        deleted_paths = [
            relpath
            for relpath in self.iter_files(self.root)
            if relpath not in self.files
        ]

        # Update workdir to match entries
        for entry in self.files.values():
            entry.write(self.root)

        # Delete files that do not correspond to entries
        for path in deleted_paths:
            (self.root / path).unlink()
        self.clean_empty_dirs(deleted_paths)

    def pull(
        self,
        repo: RemoteTaskConfigurationRepository,
        logger: logging.Logger,
    ) -> PullStats:
        """
        Pull changes from another repository.

        :param repo: repository to pull from
        :param logger: logger to use to give user feedback about what changed
        """
        return self._pull(repo, logger, preserve_deleted=True)

    @classmethod
    def iter_files(cls, root: Path) -> Generator[Path]:
        """Generate a sequence of relative paths for all yaml files in path."""
        for cur_root, _dirs, files in root.walk():
            for name in files:
                if not name.endswith(".yaml"):
                    continue
                yield (cur_root / name).relative_to(root)

    @classmethod
    def is_checkout(cls, root: Path) -> bool:
        """Check if there is a task configuration checkout at path."""
        return (root / "MANIFEST").exists()

    @classmethod
    def read_manifest(cls, root: Path) -> Manifest:
        """Load manifest data from the repo at the given path."""
        path = root / "MANIFEST"
        try:
            with path.open() as fd:
                data = yaml.safe_load(fd)
            return Manifest(**data)
        except Exception as exc:
            exc.add_note(f"manifest: {path}")
            raise

    @classmethod
    def from_path(cls, root: Path) -> "LocalTaskConfigurationRepository":
        """Instantiate a local repository from a path."""
        if GitTaskConfigurationRepository.is_git(root):
            return GitTaskConfigurationRepository(root)
        else:
            return LocalTaskConfigurationRepository(root)


class GitTaskConfigurationRepository(LocalTaskConfigurationRepository):
    """LocalTaskConfigurationRepository which is a git repository."""

    def git(
        self, args: list[str], check: bool = True
    ) -> subprocess.CompletedProcess[str]:
        """Run a git command in the repository and return its output."""
        cmd = ["git", "-C", self.root.as_posix(), *args]
        res = subprocess.run(cmd, capture_output=True, text=True)
        if check and res.returncode != 0:
            try:
                res.check_returncode()
            except Exception as e:
                e.add_note("stderr:\n" + res.stderr)
                raise
        return res

    @override
    def is_dirty(self) -> bool:
        res = self.git(["status", "--porcelain"], check=True)
        for line in res.stdout.splitlines():
            if line.endswith("/"):
                return True
            elif line.endswith(".yaml"):
                return True
            elif line[3:] == "MANIFEST":
                return True
        return False

    @override
    def git_commit(self) -> str | None:
        res = self.git(["rev-parse", "HEAD"], check=False)
        if res.returncode == 0:
            return res.stdout.strip()
        # rev-parse may fail if the git repo has just been initialized
        return None

    @override
    def has_commit(self, commit: str) -> bool:
        res = self.git(
            ["merge-base", "--is-ancestor", commit, "HEAD"], check=False
        )
        if res.returncode == 0:
            return True
        return False

    @override
    def pull(
        self,
        repo: RemoteTaskConfigurationRepository,
        logger: logging.Logger,
    ) -> PullStats:
        if self.is_dirty():
            raise InvalidRepository(
                "git working directory is dirty: please commit changes"
                " before pulling from debusine"
            )

        return self._pull(repo, logger, preserve_deleted=False)

    @classmethod
    def is_git(cls, path: Path) -> bool:
        """
        Check if the given path contains a git repository.

        If git is not installed, it returns False.
        """
        if (git := shutil.which("git")) is None:
            return False
        res = subprocess.run(
            [git, "-C", path.as_posix(), "rev-parse"], capture_output=True
        )
        return res.returncode == 0



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command list_notification_channels."""

from typing import override

from django.test import TestCase
from rich.pretty import pretty_repr

from debusine.db.models import NotificationChannel
from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.server.management.management_utils import sort_keys


class ListNotificationChannelsTests(TabularOutputTests, TestCase):
    """Tests for the list_notification_channels command."""

    @override
    def setUp(self) -> None:
        """Set up new notification channel for the test."""
        super().setUp()
        data = {
            "from": "sender@debusine.example.org",
            "to": ["recipient@example.com"],
        }
        self.notification_channel = NotificationChannel.objects.create(
            name="lts", method=NotificationChannel.Methods.EMAIL, data=data
        )

    def test_list_notification_channels(self) -> None:
        """list_notification_channels print the channels."""
        with self.assertPrintsTable() as output:
            _, stderr, exit_code = call_command("list_notification_channels")
        self.assertEqual(output.col(0), [self.notification_channel.name])
        self.assertEqual(output.col(1), [self.notification_channel.method])
        self.assertEqual(
            output.col(2),
            [pretty_repr(sort_keys(self.notification_channel.data))],
        )

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common utilities for Django-based packages."""

import os
import shutil
import sys
import tempfile
from pathlib import Path
from typing import Any

from django.conf import settings


def in_test_suite() -> bool:
    """Check whether we're running in the test suite."""
    # `"PYTEST_VERSION" in os.environ` would be a cleaner and simpler way to
    # detect whether pytest is running, but that requires pytest >= 8.2.
    return (
        sys.argv[1:2] == ["test"]
        or "pytest" in sys.argv[0].split(os.sep)
        or "PYTEST_XDIST_WORKER" in os.environ
    )


def setup_temp_data_directories() -> None:
    """
    Create and configure temporary directories for use by the test suite.

    Do not pollute the development/production directories, such as the store.
    Called through from LOGGING_CONFIG, after settings are fully processed.
    """
    tempdir = tempfile.mkdtemp(prefix="debusine-testsuite-data-")
    settings.DEBUSINE_DATA_PATH = tempdir
    settings.DEBUSINE_CACHE_DIRECTORY = os.path.join(tempdir, 'cache')
    settings.DEBUSINE_TEMPLATE_DIRECTORY = os.path.join(tempdir, 'templates')
    settings.DEBUSINE_UPLOAD_DIRECTORY = os.path.join(tempdir, 'uploads')
    settings.DEBUSINE_STORE_DIRECTORY = os.path.join(tempdir, 'store')
    settings.DEBUSINE_TOKENS_DIRECTORY = os.path.join(tempdir, 'tokens')
    os.mkdir(settings.DEBUSINE_CACHE_DIRECTORY)
    os.mkdir(settings.DEBUSINE_TEMPLATE_DIRECTORY)
    os.mkdir(settings.DEBUSINE_UPLOAD_DIRECTORY)
    os.mkdir(settings.DEBUSINE_STORE_DIRECTORY)
    os.mkdir(settings.DEBUSINE_TOKENS_DIRECTORY, mode=0o700)

    if hasattr(settings, "DEBUSINE_LOG_DIRECTORY"):
        settings.DEBUSINE_LOG_DIRECTORY = os.path.join(tempdir, 'logs')
        os.mkdir(settings.DEBUSINE_LOG_DIRECTORY)

    if settings.TEMPLATES:
        settings.TEMPLATES[0]["DIRS"] = [settings.DEBUSINE_TEMPLATE_DIRECTORY]

    # We could send signals for tests that cache settings, and restore them:
    # https://docs.djangoproject.com/en/3.2/ref/signals/#django.test.signals.setting_changed
    # but we're early in the start-up process, this complexifies the code,
    # and it's hard to assess the side effects, so let's not.


def cleanup_temp_directory() -> None:
    """Cleanup a temporary DEBUSINE_DATA_PATH that we created for tests."""
    data_path = Path(settings.DEBUSINE_DATA_PATH)

    temp_root = Path(tempfile.gettempdir())
    if data_path.is_relative_to(temp_root):
        shutil.rmtree(data_path)
    else:
        raise AssertionError(
            f"Not deleting test directory {str(data_path)!r} "
            f"(outside {str(temp_root)!r}).\n",
        )


def test_data_override(config: dict[str, Any]) -> None:
    """
    Setup temp data directories for testing.

    This is called through LOGGING_CONFIG, because that hooks is
    invoked after the settings are processed (including
    .__init__.compute_default_settings()), right before the
    application starts.
    """
    import logging.config

    from django.conf import settings

    test_mode = getattr(settings, "TEST_MODE", False)
    if test_mode:  # pragma: no branch
        # Replace and create DEBUSINE_*_DIRECTORY
        setup_temp_data_directories()

        if hasattr(settings, "DEBUSINE_LOG_DIRECTORY"):
            # Divert log handlers to new temp DEBUSINE_LOG_DIRECTORY
            for handler in config["handlers"].values():
                if "filename" in handler:
                    handler["filename"] = os.path.join(
                        settings.DEBUSINE_LOG_DIRECTORY,
                        os.path.basename(handler["filename"]),
                    )

        # Avoid writing test logs to journald
        if "journal" in config["handlers"]:
            config["handlers"]["journal"] = {
                "class": "logging.NullHandler",
            }

    # default logging config handler
    logging.config.dictConfig(config)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for log filters."""

import logging
from typing import override
from unittest.mock import Mock

from django.db.utils import OperationalError
from django.test import SimpleTestCase

from debusine.django.log_filters import SuppressDatabaseErrors


class TestSuppressDatabaseErrors(SimpleTestCase):
    """Tests for the SuppressDatabaseErrors logging filter."""

    @override
    def setUp(self) -> None:
        self.filter = SuppressDatabaseErrors()
        self.record = Mock(spec=logging.LogRecord)

    def test_suppress_operational_error(self) -> None:
        """OperationalError should be suppressed."""
        self.record.exc_info = (OperationalError, OperationalError(), None)
        self.assertFalse(self.filter.filter(self.record))

    def test_allow_other_error(self) -> None:
        """Other errors like ValueError should be allowed."""
        self.record.exc_info = (ValueError, ValueError(), None)
        self.assertTrue(self.filter.filter(self.record))

    def test_allow_no_exception_type(self) -> None:
        """Records with exc_info but no exception type should be allowed."""
        self.record.exc_info = (None, None, None)
        self.assertTrue(self.filter.filter(self.record))

    def test_allow_no_exc_info(self) -> None:
        """Records without exc_info should be allowed."""
        self.record.exc_info = None
        self.assertTrue(self.filter.filter(self.record))

    def test_suppress_operational_error_subclass(self) -> None:
        """Subclasses of OperationalError should also be suppressed."""

        class MyOperationalError(OperationalError):
            pass

        self.record.exc_info = (MyOperationalError, MyOperationalError(), None)
        self.assertFalse(self.filter.filter(self.record))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Cloud Worker Pools."""

from debusine.server.worker_pools.amazon_ec2 import AWSEC2WorkerPool
from debusine.server.worker_pools.base import (
    WorkerPoolInterface,
    provider_interface,
)
from debusine.server.worker_pools.dummy import DummyWorkerPool
from debusine.server.worker_pools.hetzner import HetznerCloudWorkerPool
from debusine.server.worker_pools.models import (
    AWSEC2InstanceMarketType,
    AWSEC2InstanceRequirements,
    AWSEC2LaunchTemplate,
    AWSEC2NetworkInterface,
    AWSEC2WorkerPoolSpecification,
    AWSInstanceRequirementFilterEnum,
    DummyWorkerPoolSpecification,
    HetznerCloudWorkerPoolSpecification,
    MinMaxRequirements,
    ScopeWorkerPoolLimits,
    WorkerPoolLimits,
    WorkerPoolSpecifications,
    worker_pool_specifications_model,
)

__all__ = [
    "AWSEC2InstanceMarketType",
    "AWSEC2InstanceRequirements",
    "AWSEC2LaunchTemplate",
    "AWSEC2NetworkInterface",
    "AWSEC2WorkerPool",
    "AWSEC2WorkerPoolSpecification",
    "AWSInstanceRequirementFilterEnum",
    "DummyWorkerPool",
    "DummyWorkerPoolSpecification",
    "HetznerCloudWorkerPool",
    "HetznerCloudWorkerPoolSpecification",
    "MinMaxRequirements",
    "ScopeWorkerPoolLimits",
    "WorkerPoolInterface",
    "WorkerPoolLimits",
    "WorkerPoolSpecifications",
    "provider_interface",
    "worker_pool_specifications_model",
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine CLI commands."""

from debusine.client.commands.base import Command

# Commandss need to be imported in order to be available to cli.py
# They are registered via Command.__init_subclass__.
#
# Command groups are shown in command line help in the order they have been
# registered, so header sorting is disabled here.

# isort: off
import debusine.client.commands.setup  # noqa: F401
import debusine.client.commands.workspaces  # noqa: F401
import debusine.client.commands.collections  # noqa: F401
import debusine.client.commands.collection_relations  # noqa: F401
import debusine.client.commands.workflow_templates  # noqa: F401
import debusine.client.commands.workflows  # noqa: F401
import debusine.client.commands.work_requests  # noqa: F401
import debusine.client.commands.artifacts  # noqa: F401
import debusine.client.commands.assets  # noqa: F401
import debusine.client.commands.task_config  # noqa: F401
import debusine.client.commands.archives  # noqa: F401

# isort: on

__all__ = ["Command"]
